from typing import Any
from typing import Callable
from typing import Generic
from typing import Iterable
from typing import Iterator
from typing import Optional
from typing import overload
from typing import Set
from typing import Tuple as _TypingTuple
from typing import Type
from typing import TypeVar
from typing import Union

from typing_extensions import Protocol

from . import operators
from . import roles
from . import sqltypes
from . import type_api
from .annotation import Annotated
from .annotation import SupportsWrappingAnnotations
from .base import ColumnSet
from .base import Executable
from .base import Immutable
from .base import SingletonConstant
from .compiler import SQLCompiler
from .functions import FunctionElement
from .schema import ForeignKey
from .selectable import TextualSelect
from .traversals import HasCopyInternals
from .traversals import MemoizedHasCacheKey
from .visitors import Traversible
from .. import util
from ..engine import Connection
from ..engine import Dialect
from ..engine import Engine
from ..util import HasMemoized
from ..util import langhelpers

_T = TypeVar("_T")
_T_co = TypeVar("_T_co", covariant=True)
_U_co = TypeVar("_U_co", covariant=True)

_TE = TypeVar("_TE", bound=type_api.TypeEngine[Any])
_OTE = TypeVar("_OTE", bound=type_api.TypeEngine[Any])
_CLE = TypeVar("_CLE", bound=ClauseElement)
_COLE = TypeVar("_COLE", bound=ColumnElement[Any])
_TC = TypeVar("_TC", bound=TextClause)
_CL = TypeVar("_CL", bound=ClauseList[Any])
_BCL = TypeVar("_BCL", bound=BooleanClauseList[Any])
_TU = TypeVar("_TU", bound=_TypingTuple[Any])
_SL = TypeVar("_SL", bound=Slice)
_GE = TypeVar("_GE", bound=GroupedElement)
_FF = TypeVar("_FF", bound=FunctionFilter[Any])
_TYC = TypeVar("_TYC", bound=TypeCoerce[Any, Any])
_UE = TypeVar("_UE", bound=UnaryExpression[Any])
_AB = TypeVar("_AB", bound=AsBoolean[Any])
_BE = TypeVar("_BE", bound=BinaryExpression[Any])
_QN = TypeVar("_QN", bound=quoted_name)
_TL = TypeVar("_TL", bound=_truncated_label)

def collate(expression: Any, collation: str) -> BinaryExpression[_TE]: ...
def between(
    expr: Any,
    lower_bound: Any,
    upper_bound: Any,
    symmetric: bool = ...,
) -> BinaryExpression[_TE]: ...
@overload
def literal(value: _T) -> BindParameter[type_api.TypeEngine[_T]]: ...
@overload
def literal(
    value: Any, type_: Optional[Union[_TE, Type[_TE]]]
) -> BindParameter[_TE]: ...
@overload
def outparam(key: Any) -> BindParameter[Any]: ...
@overload
def outparam(
    key: Any, type_: Optional[Union[_TE, Type[_TE]]]
) -> BindParameter[_TE]: ...
def not_(
    clause: Any,
) -> UnaryExpression[sqltypes.NullType]: ...

class ClauseElement(
    roles.SQLRole,
    SupportsWrappingAnnotations,
    MemoizedHasCacheKey,
    HasCopyInternals,
    Traversible,
):
    __visit_name__: str = ...
    supports_execution: bool = ...
    stringify_dialect: str = ...
    bind: Any = ...
    description: Any = ...
    is_clause_element: bool = ...
    is_selectable: bool = ...
    def unique_params(
        self: _CLE, *optionaldict: Any, **kwargs: Any
    ) -> _CLE: ...
    def params(self: _CLE, *optionaldict: Any, **kwargs: Any) -> _CLE: ...
    def compare(self, other: Any, **kw: Any) -> bool: ...
    def get_children(
        self, omit_attrs: Any = ..., **kw: Any
    ) -> Iterable[Any]: ...
    def self_group(self, against: Optional[Any] = ...) -> ClauseElement: ...
    def compile(
        self,
        bind: Optional[Any] = ...,
        dialect: Optional[Dialect] = ...,
        **kw: Any,
    ) -> SQLCompiler: ...
    def __invert__(self) -> Any: ...
    def __bool__(self) -> bool: ...
    def __nonzero__(self) -> bool: ...

class ColumnElement(
    roles.ColumnArgumentOrKeyRole,
    roles.StatementOptionRole,
    roles.WhereHavingRole,
    roles.BinaryElementRole,
    roles.OrderByRole,
    roles.ColumnsClauseRole,
    roles.LimitOffsetRole,
    roles.DMLColumnRole,
    roles.DDLConstraintColumnRole,
    roles.DDLExpressionRole,
    operators.ColumnOperators,
    ClauseElement,
    Generic[_TE],
):
    __visit_name__: str = ...
    primary_key: Union[bool, ColumnSet[Any]] = ...
    foreign_keys: Any = ...
    key: Any = ...
    def self_group(
        self: _COLE, against: Optional[Any] = ...
    ) -> Union[_COLE, Grouping[_TE], AsBoolean[_COLE]]: ...
    @util.memoized_property
    def type(self) -> _TE: ...
    @HasMemoized.memoized_attribute
    def comparator(self) -> type_api.TypeEngine.Comparator[_TE]: ...
    def __getattr__(self, key: Any) -> Any: ...
    def operate(
        self, op: Any, *other: Any, **kwargs: Any
    ) -> ClauseElement: ...
    def reverse_operate(
        self, op: Any, other: Any, **kwargs: Any
    ) -> ClauseElement: ...
    @property
    def expression(self: _COLE) -> _COLE: ...
    @util.memoized_property
    def base_columns(self) -> util.column_set[Any]: ...
    @util.memoized_property
    def proxy_set(self) -> util.column_set[Any]: ...
    def shares_lineage(self, othercolumn: Any) -> bool: ...
    def cast(
        self: _COLE, type_: Union[Type[_OTE], _OTE]
    ) -> Cast[_OTE, _COLE]: ...
    def label(self, name: Any) -> Label[_TE]: ...
    @util.memoized_property
    def anon_label(self) -> Any: ...
    @util.memoized_property
    def anon_key_label(self) -> Any: ...

class WrapsColumnExpression(Generic[_CLE]):
    @property
    def wrapped_column_expression(self) -> _CLE: ...
    @property
    def anon_label(self) -> Any: ...

class BindParameter(roles.InElementRole, ColumnElement[_TE]):
    __visit_name__: str = ...
    inherit_cache: bool = ...
    key: Any = ...
    unique: bool = ...
    value: Any = ...
    callable: Optional[Callable[[], Any]] = ...
    isoutparam: Any = ...
    required: bool = ...
    expanding: Any = ...
    literal_execute: Any = ...
    type: _TE = ...  # type: ignore
    def __init__(
        self,
        key: Any,
        value: Any = ...,
        type_: Optional[Union[Type[_TE], _TE]] = ...,
        unique: bool = ...,
        required: bool = ...,
        quote: Optional[bool] = ...,
        callable_: Optional[Callable[[], Any]] = ...,
        expanding: bool = ...,
        isoutparam: bool = ...,
        literal_execute: bool = ...,
        _compared_to_operator: Optional[Any] = ...,
        _compared_to_type: Optional[Any] = ...,
        _is_crud: bool = ...,
    ) -> None: ...
    @property
    def effective_value(self) -> Any: ...

class TypeClause(ClauseElement, Generic[_TE]):
    __visit_name__: str = ...
    type: _TE = ...
    def __init__(self, type_: _TE) -> None: ...

class TextClause(
    roles.DDLConstraintColumnRole,
    roles.DDLExpressionRole,
    roles.StatementOptionRole,
    roles.WhereHavingRole,
    roles.OrderByRole,
    roles.FromClauseRole,
    roles.SelectStatementRole,
    roles.BinaryElementRole,
    roles.InElementRole,
    Executable,
    ClauseElement,
):
    __visit_name__: str = ...
    def __and__(
        self: _TC, other: _CLE
    ) -> BooleanClauseList[Union[_TC, _CLE]]: ...
    key: Any = ...
    text: Any = ...
    def __init__(self, text: util.text_type, bind: Optional[Any] = ...): ...
    @classmethod
    def _create_text(
        cls,
        text: util.text_type,
        bind: Optional[Union[Engine, Connection]] = ...,
    ) -> TextClause: ...
    def bindparams(self: _TC, *binds: Any, **names_to_values: Any) -> _TC: ...
    def columns(self, *cols: Any, **types: Any) -> TextualSelect: ...
    @property
    def type(self) -> sqltypes.NullType: ...
    @property
    def comparator(self) -> sqltypes.NullType.Comparator: ...
    def self_group(
        self: _TC, against: Optional[Any] = ...
    ) -> Union[Grouping[sqltypes.NullType], _TC]: ...

class Null(
    SingletonConstant,
    roles.ConstExprRole,
    ColumnElement[sqltypes.NullType],
):
    __visit_name__: str = ...
    @classmethod
    def _instance(cls) -> Null: ...

class False_(
    SingletonConstant,
    roles.ConstExprRole,
    ColumnElement[sqltypes.Boolean],
):
    __visit_name__: str = ...
    @classmethod
    def _instance(cls) -> False_: ...

class True_(
    SingletonConstant,
    roles.ConstExprRole,
    ColumnElement[sqltypes.Boolean],
):
    __visit_name__: str = ...
    @classmethod
    def _instance(cls) -> True_: ...

class ClauseList(
    roles.InElementRole,
    roles.OrderByRole,
    roles.ColumnsClauseRole,
    roles.DMLColumnRole,
    ClauseElement,
    Generic[_CLE],
):
    __visit_name__: str = ...
    operator: Any = ...
    group: Any = ...
    group_contents: Any = ...
    clauses: Any = ...
    def __init__(self, *clauses: _CLE, **kwargs: Any) -> None: ...
    def __iter__(self) -> Iterator[_CLE]: ...
    def __len__(self) -> int: ...
    def append(self, clause: _CLE) -> None: ...
    def self_group(
        self: _CL, against: Optional[Any] = ...
    ) -> Union[_CL, Grouping[sqltypes.NullType]]: ...

class BooleanClauseList(
    ClauseList[_CLE], ColumnElement[sqltypes.Boolean], Generic[_CLE]
):
    __visit_name__: str = ...
    inherit_cache: bool = ...
    def __init__(self, *arg: Any, **kw: Any) -> None: ...
    @classmethod
    def and_(cls, *clauses: _CLE) -> BooleanClauseList[_CLE]: ...
    @classmethod
    def or_(cls, *clauses: _CLE) -> BooleanClauseList[_CLE]: ...
    def self_group(  # type: ignore[override]
        self: _BCL, against: Optional[Any] = ...
    ) -> Union[_BCL, Grouping[sqltypes.Boolean]]: ...

and_ = BooleanClauseList.and_
or_ = BooleanClauseList.or_

class Tuple(ClauseList[_CLE], ColumnElement[_TE]):
    __visit_name__: str = ...
    type: sqltypes.TupleType[_TE] = ...  # type: ignore[assignment]
    def __init__(self, *clauses: _CLE, **kw: Any) -> None: ...
    def self_group(self: _TU, against: Optional[Any] = ...) -> _TU: ...  # type: ignore

class Case(ColumnElement[_TE]):
    __visit_name__: str = ...
    value: Any = ...
    type: Any = ...
    whens: Any = ...
    else_: Any = ...
    def __init__(self, *whens: Any, **kw: Any) -> None: ...

@overload
def literal_column(
    text: str, type_: None = ...
) -> ColumnClause[sqltypes.NullType]: ...
@overload
def literal_column(
    text: str, type_: Union[_TE, Type[_TE]]
) -> ColumnClause[_TE]: ...

class Cast(WrapsColumnExpression[_CLE], ColumnElement[_TE], Generic[_TE, _CLE]):  # type: ignore[misc]
    __visit_name__: str = ...
    type: _TE = ...  # type: ignore[assignment]
    clause: _CLE = ...
    typeclause: TypeClause[_TE] = ...
    def __init__(
        self, expression: _CLE, type_: Union[_TE, Type[_TE]]
    ) -> None: ...
    @property
    def wrapped_column_expression(self) -> _CLE: ...

class TypeCoerce(WrapsColumnExpression[_CLE], ColumnElement[_TE], Generic[_TE, _CLE]):  # type: ignore[misc]
    __visit_name__: str = ...
    type: _TE = ...  # type: ignore[assignment]
    clause: _CLE = ...
    def __init__(
        self, expression: _CLE, type_: Union[_TE, Type[_TE]]
    ) -> None: ...
    @HasMemoized.memoized_attribute
    def typed_expression(self) -> Any: ...
    @property
    def wrapped_column_expression(self) -> _CLE: ...
    def self_group(
        self: _TYC, against: Optional[Any] = ...
    ) -> Union[_TYC, Grouping[Any]]: ...

class Extract(ColumnElement[sqltypes.Integer]):
    __visit_name__: str = ...
    type: sqltypes.Integer = ...  # type: ignore[assignment]
    field: Any = ...
    expr: Any = ...
    def __init__(self, field: Any, expr: Any, **kwargs: Any) -> None: ...

class _label_reference(ColumnElement[sqltypes.NullType]):
    __visit_name__: str = ...
    element: Any = ...
    def __init__(self, element: Any) -> None: ...

class _textual_label_reference(ColumnElement[sqltypes.NullType]):
    __visit_name__: str = ...
    element: Any = ...
    def __init__(self, element: Any) -> None: ...

class UnaryExpression(ColumnElement[_TE]):
    __visit_name__: str = ...
    operator: Any = ...
    modifier: Any = ...
    element: Any = ...
    type: _TE = ...  # type: ignore[assignment]
    wraps_column_expression: Any = ...
    @overload
    def __init__(
        self,
        element: Any,
        *,
        operator: Any = ...,
        modifier: Any = ...,
        type_: Union[_TE, Type[_TE]],
        wraps_column_expression: bool = ...,
    ) -> None: ...
    @overload
    def __init__(
        self,
        element: Any,
        operator: Any,
        modifier: Any,
        type_: Union[_TE, Type[_TE]],
        wraps_column_expression: bool = ...,
    ) -> None: ...
    @overload
    def __init__(
        self: UnaryExpression[sqltypes.NullType],
        element: Any,
        operator: Optional[Any] = ...,
        modifier: Optional[Any] = ...,
        type_: None = ...,
        wraps_column_expression: bool = ...,
    ) -> None: ...
    @classmethod
    def _create_nulls_first(
        cls, column: Any
    ) -> UnaryExpression[sqltypes.NullType]: ...
    @classmethod
    def _create_nulls_last(
        cls, column: Any
    ) -> UnaryExpression[sqltypes.NullType]: ...
    @classmethod
    def _create_desc(
        cls, column: Any
    ) -> UnaryExpression[sqltypes.NullType]: ...
    @classmethod
    def _create_asc(
        cls, column: Any
    ) -> UnaryExpression[sqltypes.NullType]: ...
    @classmethod
    def _create_distinct(cls, column: Any) -> UnaryExpression[_OTE]: ...
    def self_group(
        self: _UE, against: Optional[Any] = ...
    ) -> Union[_UE, Grouping[_TE]]: ...

class CollectionAggregate(UnaryExpression[_TE]):
    @classmethod
    def _create_any(
        cls, expr: Any
    ) -> CollectionAggregate[sqltypes.NullType]: ...
    @classmethod
    def _create_all(
        cls, expr: Any
    ) -> CollectionAggregate[sqltypes.NullType]: ...
    def operate(
        self, op: Any, *other: Any, **kwargs: Any
    ) -> ClauseElement: ...
    def reverse_operate(self, op: Any, other: Any, **kwargs: Any) -> None: ...  # type: ignore[override]

class AsBoolean(  # type: ignore[misc]
    WrapsColumnExpression[_CLE],
    UnaryExpression[sqltypes.Boolean],
    Generic[_CLE],
):
    inherit_cache: bool = ...
    element: _CLE = ...
    type: sqltypes.Boolean = ...
    operator: Any = ...
    negate: Any = ...
    modifier: Any = ...
    wraps_column_expression: bool = ...
    def __init__(self, element: _CLE, operator: Any, negate: Any) -> None: ...
    @property
    def wrapped_column_expression(self) -> _CLE: ...
    def self_group(self: _AB, against: Optional[Any] = ...) -> _AB: ...  # type: ignore[override]

class BinaryExpression(ColumnElement[_TE]):
    __visit_name__: str = ...
    left: Any = ...
    right: Any = ...
    operator: Any = ...
    type: _TE = ...  # type: ignore[assignment]
    negate: Any = ...
    modifiers: Any = ...
    @overload
    def __init__(
        self,
        left: Any,
        right: Any,
        operator: Any,
        type_: Union[_TE, Type[_TE]],
        negate: Optional[Any] = ...,
        modifiers: Optional[Any] = ...,
    ) -> None: ...
    @overload
    def __init__(
        self: BinaryExpression[sqltypes.NullType],
        left: Any,
        right: Any,
        operator: Any,
        type_: None = ...,
        negate: Optional[Any] = ...,
        modifiers: Optional[Any] = ...,
    ) -> None: ...
    def __bool__(self) -> bool: ...
    def __nonzero__(self) -> bool: ...
    @property
    def is_comparison(self) -> bool: ...
    def self_group(
        self: _BE, against: Optional[Any] = ...
    ) -> Union[_BE, Grouping[_TE]]: ...

class Slice(ColumnElement[sqltypes.NullType]):
    __visit_name__: str = ...
    start: Any = ...
    stop: Any = ...
    step: Any = ...
    type: sqltypes.NullType = ...  # type: ignore[assignment]
    def __init__(
        self,
        start: ClauseElement,
        stop: ClauseElement,
        step: ClauseElement,
        _name: Optional[Any] = ...,
    ) -> None: ...
    def self_group(self: _SL, against: Optional[Any] = ...) -> _SL: ...

class IndexExpression(BinaryExpression[_TE]): ...

class GroupedElement(ClauseElement):
    __visit_name__: str = ...
    def self_group(self: _GE, against: Optional[Any] = ...) -> _GE: ...

class Grouping(GroupedElement, ColumnElement[_TE]):  # type: ignore[misc]
    element: ClauseElement = ...
    type: _TE = ...  # type: ignore[assignment]
    @overload
    def __init__(self, element: ColumnElement[_TE]) -> None: ...
    @overload
    def __init__(
        self: Grouping[sqltypes.NullType], element: ClauseElement
    ) -> None: ...
    def __getattr__(self, attr: Any) -> Any: ...

RANGE_UNBOUNDED: langhelpers._symbol
RANGE_CURRENT: langhelpers._symbol

class Over(ColumnElement[_TE]):
    __visit_name__: str = ...
    order_by: Any = ...
    partition_by: Any = ...
    element: Any = ...
    range_: Any = ...
    rows: Any = ...
    def __init__(
        self,
        element: ColumnElement[_TE],
        partition_by: Optional[ClauseElement] = ...,
        order_by: Optional[ClauseElement] = ...,
        range_: Optional[Any] = ...,
        rows: Optional[Any] = ...,
    ) -> None: ...
    def __reduce__(self) -> Any: ...
    @util.memoized_property
    def type(self) -> _TE: ...

class _WithinGroupTypeElement(Protocol[_T_co, _U_co]):
    @property
    def type(self) -> _T_co: ...
    def within_group_type(self, within_group: Any) -> _U_co: ...

class WithinGroup(ColumnElement[_TE]):
    __visit_name__: str = ...
    order_by: ClauseList[ColumnElement[Any]] = ...
    element: _WithinGroupTypeElement[Any, Any] = ...
    @overload
    def __init__(
        self: WithinGroup[_TE],
        element: _WithinGroupTypeElement[_TE, None],
        *order_by: _COLE,
    ) -> None: ...
    @overload
    def __init__(
        self: WithinGroup[_TE],
        element: _WithinGroupTypeElement[Any, _TE],
        *order_by: _COLE,
    ) -> None: ...
    def over(
        self,
        partition_by: Optional[Any] = ...,
        order_by: Optional[Any] = ...,
        range_: Optional[Any] = ...,
        rows: Optional[Any] = ...,
    ) -> Over[_TE]: ...
    @util.memoized_property
    def type(self) -> _TE: ...

class FunctionFilter(ColumnElement[_TE]):
    __visit_name__: str = ...
    criterion: Any = ...
    func: FunctionElement[_TE] = ...
    def __init__(
        self, func: FunctionElement[_TE], *criterion: ClauseElement
    ) -> None: ...
    def filter(self: _FF, *criterion: ClauseElement) -> _FF: ...
    def over(
        self,
        partition_by: Optional[Any] = ...,
        order_by: Optional[Any] = ...,
        range_: Optional[Any] = ...,
        rows: Optional[Any] = ...,
    ) -> Over[_TE]: ...
    def self_group(
        self: _FF, against: Optional[Any] = ...
    ) -> Union[_FF, Grouping[_TE]]: ...
    @util.memoized_property
    def type(self) -> _TE: ...

class Label(roles.LabeledColumnExprRole, ColumnElement[_TE]):
    __visit_name__: str = ...
    name: str = ...
    key: Any = ...
    @overload
    def __init__(
        self, name: Any, element: ColumnElement[_TE], type_: None = ...
    ) -> None: ...
    @overload
    def __init__(
        self,
        name: str,
        element: ColumnElement[Any],
        type_: Union[_TE, Type[_TE]],
    ) -> None: ...
    def __reduce__(self) -> Any: ...
    @util.memoized_property
    def type(self) -> _TE: ...
    @HasMemoized.memoized_attribute
    def element(self) -> ClauseElement: ...
    def self_group(self, against: Optional[Any] = ...) -> ClauseElement: ...  # type: ignore[override]
    @property
    def primary_key(self) -> Union[bool, ColumnSet[Any]]: ...  # type: ignore[override]
    @property
    def foreign_keys(self) -> Set[ForeignKey]: ...

class NamedColumn(ColumnElement[_TE]):
    is_literal: bool = ...
    table: Any = ...
    @util.memoized_property
    def description(self) -> util.text_type: ...

class ColumnClause(
    roles.DDLReferredColumnRole,
    roles.LabeledColumnExprRole,
    roles.StrAsPlainColumnRole,
    Immutable,
    NamedColumn[_TE],
):
    table: Any = ...
    is_literal: bool = ...
    __visit_name__: str = ...
    onupdate: Any = ...
    default: Any = ...
    server_default: Any = ...
    server_onupdate: Any = ...
    key: Any = ...
    type: Any = ...
    @overload
    def __init__(
        self: ColumnClause[sqltypes.NullType],
        text: str,
        type_: None = ...,
        is_literal: bool = ...,
        _selectable: Optional[Any] = ...,
    ) -> None: ...
    @overload
    def __init__(
        self,
        text: str,
        type_: Union[_TE, Type[_TE]],
        is_literal: bool = ...,
        _selectable: Optional[Any] = ...,
    ) -> None: ...
    def get_children(self, column_tables: bool = ..., **kw: Any) -> Iterable[Any]: ...  # type: ignore[override]

class TableValuedColumn(NamedColumn[_TE]):
    __visit_name__: str = ...
    scalar_alias: Any = ...
    key: Any = ...
    type: Any = ...
    def __init__(self, scalar_alias: ClauseElement, type_: _TE) -> None: ...

class CollationClause(ColumnElement[_TE]):
    __visit_name__: str = ...
    collation: str = ...
    def __init__(self, collation: str) -> None: ...

class _IdentifiedClause(Executable, ClauseElement):
    __visit_name__: str = ...
    ident: Any = ...
    def __init__(self, ident: Any) -> None: ...

class SavepointClause(_IdentifiedClause):
    __visit_name__: str = ...

class RollbackToSavepointClause(_IdentifiedClause):
    __visit_name__: str = ...

class ReleaseSavepointClause(_IdentifiedClause):
    __visit_name__: str = ...

class quoted_name(util.MemoizedSlots, util.text_type):
    quote: Any = ...
    def __new__(cls: Type[_QN], value: Any, quote: Any) -> _QN: ...
    def __reduce__(self) -> Any: ...

class AnnotatedColumnElement(Annotated):
    def __init__(self, element: Any, values: Any) -> None: ...
    @util.memoized_property
    def name(self) -> str: ...
    @util.memoized_property
    def table(self) -> Any: ...
    @util.memoized_property
    def key(self) -> Any: ...
    @util.memoized_property
    def info(self) -> Any: ...
    @util.memoized_property
    def anon_label(self) -> Any: ...

class _truncated_label(quoted_name):
    def __reduce__(self) -> Any: ...
    def apply_map(self: _TL, map_: Any) -> _TL: ...

class conv(_truncated_label): ...

class _anonymous_label(_truncated_label):
    @classmethod
    def safe_construct(
        cls,
        seed: int,
        body: str,
        enclosing_label: Optional[_anonymous_label] = ...,
    ) -> _anonymous_label: ...
    def __add__(self, other: Any) -> _anonymous_label: ...
    def __radd__(self, other: Any) -> _anonymous_label: ...
    def apply_map(self, map_: Any) -> Union[quoted_name, _anonymous_label]: ...  # type: ignore[override]



# fmt: off
from typing import Any

from . import attributes as attributes
from . import evaluator as evaluator
from . import loading as loading
from . import sync as sync
from .base import state_str as state_str
from .. import future as future
from .. import sql as sql
from .. import util as util
from ..sql import coercions as coercions
from ..sql import expression as expression
from ..sql import operators as operators
from ..sql import roles as roles
from ..sql import select as select
from ..sql.base import CompileState as CompileState
from ..sql.base import Options as Options
from ..sql.dml import DeleteDMLState as DeleteDMLState
from ..sql.dml import UpdateDMLState as UpdateDMLState
from ..sql.elements import BooleanClauseList as BooleanClauseList
from ..sql.selectable import LABEL_STYLE_TABLENAME_PLUS_COL as LABEL_STYLE_TABLENAME_PLUS_COL
# fmt: on

def save_obj(
    base_mapper: Any, states: Any, uowtransaction: Any, single: bool = ...
) -> None: ...
def post_update(
    base_mapper: Any, states: Any, uowtransaction: Any, post_update_cols: Any
) -> None: ...
def delete_obj(base_mapper: Any, states: Any, uowtransaction: Any) -> None: ...

class BulkUDCompileState(CompileState):
    class default_update_options(Options): ...
    @classmethod
    def orm_pre_session_exec(
        cls,
        session: Any,
        statement: Any,
        params: Any,
        execution_options: Any,
        bind_arguments: Any,
        is_reentrant_invoke: Any,
    ): ...
    @classmethod
    def orm_setup_cursor_result(
        cls,
        session: Any,
        statement: Any,
        params: Any,
        execution_options: Any,
        bind_arguments: Any,
        result: Any,
    ): ...

class BulkORMUpdate(UpdateDMLState, BulkUDCompileState):
    mapper: Any = ...
    extra_criteria_entities: Any = ...
    @classmethod
    def create_for_statement(
        cls, statement: Any, compiler: Any, **kw: Any
    ): ...

class BulkORMDelete(DeleteDMLState, BulkUDCompileState):
    mapper: Any = ...
    extra_criteria_entities: Any = ...
    @classmethod
    def create_for_statement(
        cls, statement: Any, compiler: Any, **kw: Any
    ): ...



from typing import Any
from typing import Optional

from . import Connector as Connector
from ..util import warn_deprecated as warn_deprecated

class MxODBCConnector(Connector):
    driver: str = ...
    supports_sane_multi_rowcount: bool = ...
    supports_unicode_statements: bool = ...
    supports_unicode_binds: bool = ...
    supports_native_decimal: bool = ...
    @classmethod
    def dbapi(cls): ...
    def on_connect(self): ...
    def create_connect_args(self, url: Any): ...
    def is_disconnect(self, e: Any, connection: Any, cursor: Any): ...
    def do_executemany(
        self,
        cursor: Any,
        statement: Any,
        parameters: Any,
        context: Optional[Any] = ...,
    ) -> None: ...
    def do_execute(
        self,
        cursor: Any,
        statement: Any,
        parameters: Any,
        context: Optional[Any] = ...,
    ) -> None: ...



from typing import Any
from typing import Optional

from .base import BIT as BIT
from .base import MySQLDialect as MySQLDialect
from .base import MySQLExecutionContext as MySQLExecutionContext
from ... import util as util

class _oursqlBIT(BIT):
    def result_processor(self, dialect: Any, coltype: Any) -> None: ...

class MySQLExecutionContext_oursql(MySQLExecutionContext):
    @property
    def plain_query(self): ...

class MySQLDialect_oursql(MySQLDialect):
    driver: str = ...
    supports_unicode_binds: bool = ...
    supports_unicode_statements: bool = ...
    supports_native_decimal: bool = ...
    supports_sane_rowcount: bool = ...
    supports_sane_multi_rowcount: bool = ...
    execution_ctx_cls: Any = ...
    colspecs: Any = ...
    @classmethod
    def dbapi(cls): ...
    def do_execute(
        self,
        cursor: Any,
        statement: Any,
        parameters: Any,
        context: Optional[Any] = ...,
    ) -> None: ...
    def do_begin(self, connection: Any) -> None: ...
    def do_begin_twophase(self, connection: Any, xid: Any) -> None: ...
    def do_prepare_twophase(self, connection: Any, xid: Any) -> None: ...
    def do_rollback_twophase(
        self,
        connection: Any,
        xid: Any,
        is_prepared: bool = ...,
        recover: bool = ...,
    ) -> None: ...
    def do_commit_twophase(
        self,
        connection: Any,
        xid: Any,
        is_prepared: bool = ...,
        recover: bool = ...,
    ) -> None: ...
    def has_table(
        self, connection: Any, table_name: Any, schema: Optional[Any] = ...
    ): ...
    def get_table_options(
        self,
        connection: Any,
        table_name: Any,
        schema: Optional[Any] = ...,
        **kw: Any,
    ): ...
    def get_columns(
        self,
        connection: Any,
        table_name: Any,
        schema: Optional[Any] = ...,
        **kw: Any,
    ): ...
    def get_view_names(
        self, connection: Any, schema: Optional[Any] = ..., **kw: Any
    ): ...
    def get_table_names(
        self, connection: Any, schema: Optional[Any] = ..., **kw: Any
    ): ...
    def get_schema_names(self, connection: Any, **kw: Any): ...
    def initialize(self, connection: Any): ...
    def is_disconnect(self, e: Any, connection: Any, cursor: Any): ...
    def create_connect_args(self, url: Any): ...

dialect = MySQLDialect_oursql



from .engine import AsyncConnection as AsyncConnection
from .engine import AsyncEngine as AsyncEngine
from .engine import AsyncTransaction as AsyncTransaction
from .engine import create_async_engine as create_async_engine
from .events import AsyncConnectionEvents as AsyncConnectionEvents
from .events import AsyncSessionEvents as AsyncSessionEvents
from .result import AsyncMappingResult as AsyncMappingResult
from .result import AsyncResult as AsyncResult
from .result import AsyncScalarResult as AsyncScalarResult
from .session import AsyncSession as AsyncSession
from .session import AsyncSessionTransaction as AsyncSessionTransaction



from operator import add as add
from operator import and_ as and_
from operator import contains as contains
from operator import eq as eq
from operator import ge as ge
from operator import getitem as getitem
from operator import gt as gt
from operator import inv as inv
from operator import le as le
from operator import lshift as lshift
from operator import lt as lt
from operator import mod as mod
from operator import mul as mul
from operator import ne as ne
from operator import neg as neg
from operator import or_ as or_
from operator import rshift as rshift
from operator import sub as sub
from operator import truediv as truediv
from typing import Any
from typing import Callable
from typing import Optional
from typing import Type
from typing import TypeVar
from typing import Union

from .elements import ClauseElement
from .type_api import TypeEngine

_F = TypeVar("_F", bound=Callable[..., Any])

div = truediv

class Operators:
    def __and__(self, other: Any) -> ClauseElement: ...
    def __or__(self, other: Any) -> ClauseElement: ...
    def __invert__(self) -> ClauseElement: ...
    def op(
        self,
        opstring: Any,
        precedence: int = ...,
        is_comparison: bool = ...,
        return_type: Optional[
            Union[Type[TypeEngine[Any]], TypeEngine[Any]]
        ] = ...,
    ) -> Callable[[Any], ClauseElement]: ...
    def bool_op(
        self, opstring: Any, precedence: int = ...
    ) -> Callable[[Any], ClauseElement]: ...
    def operate(
        self, op: Any, *other: Any, **kwargs: Any
    ) -> ClauseElement: ...
    def reverse_operate(
        self, op: Any, other: Any, **kwargs: Any
    ) -> ClauseElement: ...

class custom_op:
    __name__: str = ...
    opstring: Any = ...
    precedence: int = ...
    is_comparison: bool = ...
    natural_self_precedent: bool = ...
    eager_grouping: bool = ...
    return_type: Optional[Union[Type[TypeEngine[Any]], TypeEngine[Any]]] = ...
    def __init__(
        self,
        opstring: Any,
        precedence: int = ...,
        is_comparison: bool = ...,
        return_type: Optional[
            Union[Type[TypeEngine[Any]], TypeEngine[Any]]
        ] = ...,
        natural_self_precedent: bool = ...,
        eager_grouping: bool = ...,
    ) -> None: ...
    def __eq__(self, other: Any) -> bool: ...
    def __hash__(self) -> int: ...
    def __call__(self, left: Any, right: Any, **kw: Any) -> ClauseElement: ...

class ColumnOperators(Operators):
    timetuple: Any = ...
    def __lt__(self, other: Any) -> ClauseElement: ...
    def __le__(self, other: Any) -> ClauseElement: ...
    def __eq__(self, other: Any) -> ClauseElement: ...  # type: ignore[override]
    def __ne__(self, other: Any) -> ClauseElement: ...  # type: ignore[override]
    def is_distinct_from(self, other: Any) -> ClauseElement: ...
    def is_not_distinct_from(self, other: Any) -> ClauseElement: ...
    def isnot_distinct_from(self, other: Any) -> ClauseElement: ...
    def __gt__(self, other: Any) -> ClauseElement: ...
    def __ge__(self, other: Any) -> ClauseElement: ...
    def __neg__(self) -> ClauseElement: ...
    def __contains__(self, other: Any) -> ClauseElement: ...
    def __getitem__(self, index: Any) -> ClauseElement: ...
    def __lshift__(self, other: Any) -> ClauseElement: ...
    def __rshift__(self, other: Any) -> ClauseElement: ...
    def concat(self, other: Any) -> ClauseElement: ...
    def like(
        self, other: Any, escape: Optional[Any] = ...
    ) -> ClauseElement: ...
    def ilike(
        self, other: Any, escape: Optional[Any] = ...
    ) -> ClauseElement: ...
    def in_(self, other: Any) -> ClauseElement: ...
    def not_in(self, other: Any) -> ClauseElement: ...
    def notin_(self, other: Any) -> ClauseElement: ...
    def not_like(
        self, other: Any, escape: Optional[Any] = ...
    ) -> ClauseElement: ...
    def notlike(
        self, other: Any, escape: Optional[Any] = ...
    ) -> ClauseElement: ...
    def not_ilike(
        self, other: Any, escape: Optional[Any] = ...
    ) -> ClauseElement: ...
    def notilike(
        self, other: Any, escape: Optional[Any] = ...
    ) -> ClauseElement: ...
    def is_(self, other: Any) -> ClauseElement: ...
    def is_not(self, other: Any) -> ClauseElement: ...
    isnot: Any = ...
    def startswith(self, other: Any, **kwargs: Any) -> ClauseElement: ...
    def endswith(self, other: Any, **kwargs: Any) -> ClauseElement: ...
    def contains(self, other: Any, **kwargs: Any) -> ClauseElement: ...
    def match(self, other: Any, **kwargs: Any) -> ClauseElement: ...
    def regexp_match(
        self, pattern: Any, flags: Optional[Any] = ...
    ) -> ClauseElement: ...
    def regexp_replace(
        self, pattern: Any, replacement: Any, flags: Optional[Any] = ...
    ) -> ClauseElement: ...
    def desc(self) -> ClauseElement: ...
    def asc(self) -> ClauseElement: ...
    def nulls_first(self) -> ClauseElement: ...
    def nullsfirst(self) -> ClauseElement: ...
    def nulls_last(self) -> ClauseElement: ...
    def nullslast(self) -> ClauseElement: ...
    def collate(self, collation: Any) -> ClauseElement: ...
    def __radd__(self, other: Any) -> ClauseElement: ...
    def __rsub__(self, other: Any) -> ClauseElement: ...
    def __rmul__(self, other: Any) -> ClauseElement: ...
    def __rdiv__(self, other: Any) -> ClauseElement: ...
    def __rmod__(self, other: Any) -> ClauseElement: ...
    def between(
        self, cleft: Any, cright: Any, symmetric: bool = ...
    ) -> ClauseElement: ...
    def distinct(self) -> ClauseElement: ...
    def any_(self) -> ClauseElement: ...
    def all_(self) -> ClauseElement: ...
    def __add__(self, other: Any) -> ClauseElement: ...
    def __sub__(self, other: Any) -> ClauseElement: ...
    def __mul__(self, other: Any) -> ClauseElement: ...
    def __div__(self, other: Any) -> ClauseElement: ...
    def __mod__(self, other: Any) -> ClauseElement: ...
    def __truediv__(self, other: Any) -> ClauseElement: ...
    def __rtruediv__(self, other: Any) -> ClauseElement: ...

def commutative_op(fn: _F) -> _F: ...
def comparison_op(fn: _F) -> _F: ...
def from_() -> None: ...
def function_as_comparison_op() -> None: ...
def as_() -> None: ...
def exists() -> None: ...
def is_true(a: Any) -> None: ...

istrue = is_true

def is_false(a: Any) -> None: ...

isfalse = is_false

def is_distinct_from(a: Any, b: Any) -> ClauseElement: ...
def is_not_distinct_from(a: Any, b: Any) -> ClauseElement: ...

isnot_distinct_from = is_not_distinct_from

def is_(a: Any, b: Any) -> ClauseElement: ...
def is_not(a: Any, b: Any) -> ClauseElement: ...

isnot = is_not

def collate(a: Any, b: Any) -> ClauseElement: ...
def op(a: Any, opstring: str, b: Any) -> ClauseElement: ...
def like_op(a: Any, b: Any, escape: Optional[Any] = ...) -> ClauseElement: ...
def not_like_op(
    a: Any, b: Any, escape: Optional[Any] = ...
) -> ClauseElement: ...

notlike_op = not_like_op

def ilike_op(a: Any, b: Any, escape: Optional[Any] = ...) -> ClauseElement: ...
def not_ilike_op(
    a: Any, b: Any, escape: Optional[Any] = ...
) -> ClauseElement: ...

notilike_op = not_ilike_op

def between_op(
    a: Any, b: Any, c: Any, symmetric: bool = ...
) -> ClauseElement: ...
def not_between_op(
    a: Any, b: Any, c: Any, symmetric: bool = ...
) -> ClauseElement: ...

notbetween_op = not_between_op

def in_op(a: Any, b: Any) -> ClauseElement: ...
def not_in_op(a: Any, b: Any) -> ClauseElement: ...

notin_op = not_in_op

def distinct_op(a: Any) -> ClauseElement: ...
def any_op(a: Any) -> ClauseElement: ...
def all_op(a: Any) -> ClauseElement: ...
def startswith_op(
    a: Any, b: Any, escape: Optional[Any] = ..., autoescape: bool = ...
) -> ClauseElement: ...
def not_startswith_op(
    a: Any, b: Any, escape: Optional[Any] = ..., autoescape: bool = ...
) -> ClauseElement: ...

notstartswith_op = not_startswith_op

def endswith_op(
    a: Any, b: Any, escape: Optional[Any] = ..., autoescape: bool = ...
) -> ClauseElement: ...
def not_endswith_op(
    a: Any, b: Any, escape: Optional[Any] = ..., autoescape: bool = ...
) -> ClauseElement: ...

notendswith_op = not_endswith_op

def contains_op(
    a: Any, b: Any, escape: Optional[Any] = ..., autoescape: bool = ...
) -> ClauseElement: ...
def not_contains_op(
    a: Any, b: Any, escape: Optional[Any] = ..., autoescape: bool = ...
) -> ClauseElement: ...

notcontains_op = not_contains_op

def match_op(a: Any, b: Any, **kw: Any) -> ClauseElement: ...
def regexp_match_op(
    a: Any, b: Any, flags: Optional[Any] = ...
) -> ClauseElement: ...
def not_regexp_match_op(
    a: Any, b: Any, flags: Optional[Any] = ...
) -> ClauseElement: ...
def regexp_replace_op(
    a: Any, b: Any, replacement: Any, flags: Optional[Any] = ...
) -> ClauseElement: ...
def not_match_op(a: Any, b: Any, **kw: Any) -> ClauseElement: ...

notmatch_op = not_match_op

def comma_op(a: Any, b: Any) -> None: ...
def filter_op(a: Any, b: Any) -> None: ...
def concat_op(a: Any, b: Any) -> ClauseElement: ...
def desc_op(a: Any) -> ClauseElement: ...
def asc_op(a: Any) -> ClauseElement: ...
def nulls_first_op(a: Any) -> ClauseElement: ...

nullsfirst_op = nulls_first_op

def nulls_last_op(a: Any) -> ClauseElement: ...

nullslast_op = nulls_last_op

def json_getitem_op(a: Any, b: Any) -> None: ...
def json_path_getitem_op(a: Any, b: Any) -> None: ...
def is_comparison(op: Any) -> bool: ...
def is_commutative(op: Any) -> bool: ...
def is_ordering_modifier(op: Any) -> bool: ...
def is_natural_self_precedent(op: Any) -> bool: ...
def is_boolean(op: Any) -> bool: ...
def mirror(op: Any) -> Any: ...
def is_associative(op: Any) -> bool: ...
def is_precedent(operator: Any, against: Any) -> bool: ...



from typing import Any
from typing import Optional

from . import json as json
from .base import ENUM as ENUM
from .base import INTERVAL as INTERVAL
from .base import OID as OID
from .base import PGCompiler as PGCompiler
from .base import PGDialect as PGDialect
from .base import PGExecutionContext as PGExecutionContext
from .base import PGIdentifierPreparer as PGIdentifierPreparer
from .base import REGCLASS as REGCLASS
from .base import UUID as UUID
from ... import exc as exc
from ... import pool as pool
from ... import processors as processors
from ... import util as util
from ...sql import sqltypes as sqltypes
from ...util.concurrency import await_fallback as await_fallback
from ...util.concurrency import await_only as await_only

class AsyncpgTime(sqltypes.Time):
    def get_dbapi_type(self, dbapi: Any): ...

class AsyncpgDate(sqltypes.Date):
    def get_dbapi_type(self, dbapi: Any): ...

class AsyncpgDateTime(sqltypes.DateTime):
    def get_dbapi_type(self, dbapi: Any): ...

class AsyncpgBoolean(sqltypes.Boolean):
    def get_dbapi_type(self, dbapi: Any): ...

class AsyncPgInterval(INTERVAL):
    def get_dbapi_type(self, dbapi: Any): ...
    @classmethod
    def adapt_emulated_to_native(cls, interval: Any, **kw: Any): ...

class AsyncPgEnum(ENUM):
    def get_dbapi_type(self, dbapi: Any): ...

class AsyncpgInteger(sqltypes.Integer):
    def get_dbapi_type(self, dbapi: Any): ...

class AsyncpgBigInteger(sqltypes.BigInteger):
    def get_dbapi_type(self, dbapi: Any): ...

class AsyncpgJSON(json.JSON):
    def get_dbapi_type(self, dbapi: Any): ...
    def result_processor(self, dialect: Any, coltype: Any) -> None: ...

class AsyncpgJSONB(json.JSONB):
    def get_dbapi_type(self, dbapi: Any): ...
    def result_processor(self, dialect: Any, coltype: Any) -> None: ...

class AsyncpgJSONIndexType(sqltypes.JSON.JSONIndexType):
    def get_dbapi_type(self, dbapi: Any) -> None: ...

class AsyncpgJSONIntIndexType(sqltypes.JSON.JSONIntIndexType):
    def get_dbapi_type(self, dbapi: Any): ...

class AsyncpgJSONStrIndexType(sqltypes.JSON.JSONStrIndexType):
    def get_dbapi_type(self, dbapi: Any): ...

class AsyncpgJSONPathType(json.JSONPathType):
    def bind_processor(self, dialect: Any): ...

class AsyncpgUUID(UUID):
    def get_dbapi_type(self, dbapi: Any): ...
    def bind_processor(self, dialect: Any): ...
    def result_processor(self, dialect: Any, coltype: Any): ...

class AsyncpgNumeric(sqltypes.Numeric):
    def bind_processor(self, dialect: Any) -> None: ...
    def result_processor(self, dialect: Any, coltype: Any): ...

class AsyncpgREGCLASS(REGCLASS):
    def get_dbapi_type(self, dbapi: Any): ...

class AsyncpgOID(OID):
    def get_dbapi_type(self, dbapi: Any): ...

class PGExecutionContext_asyncpg(PGExecutionContext):
    def handle_dbapi_exception(self, e: Any) -> None: ...
    exclude_set_input_sizes: Any = ...
    def pre_exec(self) -> None: ...
    def create_server_side_cursor(self): ...

class PGCompiler_asyncpg(PGCompiler): ...
class PGIdentifierPreparer_asyncpg(PGIdentifierPreparer): ...

class AsyncAdapt_asyncpg_cursor:
    server_side: bool = ...
    description: Any = ...
    arraysize: int = ...
    rowcount: int = ...
    def __init__(self, adapt_connection: Any) -> None: ...
    def close(self) -> None: ...
    def execute(
        self, operation: Any, parameters: Optional[Any] = ...
    ) -> None: ...
    def executemany(self, operation: Any, seq_of_parameters: Any): ...
    def setinputsizes(self, *inputsizes: Any) -> None: ...
    def __iter__(self) -> Any: ...
    def fetchone(self): ...
    def fetchmany(self, size: Optional[Any] = ...): ...
    def fetchall(self): ...

class AsyncAdapt_asyncpg_ss_cursor(AsyncAdapt_asyncpg_cursor):
    server_side: bool = ...
    def __init__(self, adapt_connection: Any) -> None: ...
    def close(self) -> None: ...
    def __aiter__(self): ...
    async def __anext__(self) -> None: ...
    def fetchone(self): ...
    def fetchmany(self, size: Optional[Any] = ...): ...
    def fetchall(self): ...
    def executemany(self, operation: Any, seq_of_parameters: Any) -> None: ...

class AsyncAdapt_asyncpg_connection:
    await_: Any = ...
    dbapi: Any = ...
    isolation_level: str = ...
    readonly: bool = ...
    deferrable: bool = ...
    def __init__(
        self,
        dbapi: Any,
        connection: Any,
        prepared_statement_cache_size: int = ...,
    ) -> None: ...
    @property
    def autocommit(self): ...
    @autocommit.setter
    def autocommit(self, value: Any) -> None: ...
    def set_isolation_level(self, level: Any) -> None: ...
    def cursor(self, server_side: bool = ...): ...
    def rollback(self) -> None: ...
    def commit(self) -> None: ...
    def close(self) -> None: ...

class AsyncAdaptFallback_asyncpg_connection(AsyncAdapt_asyncpg_connection):
    await_: Any = ...

class AsyncAdapt_asyncpg_dbapi:
    asyncpg: Any = ...
    paramstyle: str = ...
    def __init__(self, asyncpg: Any) -> None: ...
    def connect(self, *arg: Any, **kw: Any): ...
    class Error(Exception): ...
    class Warning(Exception): ...
    class InterfaceError(Error): ...
    class DatabaseError(Error): ...
    class InternalError(DatabaseError): ...
    class OperationalError(DatabaseError): ...
    class ProgrammingError(DatabaseError): ...
    class IntegrityError(DatabaseError): ...
    class DataError(DatabaseError): ...
    class NotSupportedError(DatabaseError): ...
    class InternalServerError(InternalError): ...
    class InvalidCachedStatementError(NotSupportedError):
        def __init__(self, message: Any) -> None: ...
    def Binary(self, value: Any): ...
    STRING: Any = ...
    TIMESTAMP: Any = ...
    TIMESTAMP_W_TZ: Any = ...
    TIME: Any = ...
    DATE: Any = ...
    INTERVAL: Any = ...
    NUMBER: Any = ...
    FLOAT: Any = ...
    BOOLEAN: Any = ...
    INTEGER: Any = ...
    BIGINTEGER: Any = ...
    BYTES: Any = ...
    DECIMAL: Any = ...
    JSON: Any = ...
    JSONB: Any = ...
    ENUM: Any = ...
    UUID: Any = ...
    BYTEA: Any = ...
    DATETIME: Any = ...
    BINARY: Any = ...

class PGDialect_asyncpg(PGDialect):
    driver: str = ...
    supports_unicode_statements: bool = ...
    supports_server_side_cursors: bool = ...
    supports_unicode_binds: bool = ...
    default_paramstyle: str = ...
    supports_sane_multi_rowcount: bool = ...
    execution_ctx_cls: Any = ...
    statement_compiler: Any = ...
    preparer: Any = ...
    use_setinputsizes: bool = ...
    use_native_uuid: bool = ...
    colspecs: Any = ...
    is_async: bool = ...
    @classmethod
    def dbapi(cls): ...
    def set_isolation_level(self, connection: Any, level: Any) -> None: ...
    def set_readonly(self, connection: Any, value: Any) -> None: ...
    def get_readonly(self, connection: Any): ...
    def set_deferrable(self, connection: Any, value: Any) -> None: ...
    def get_deferrable(self, connection: Any): ...
    def create_connect_args(self, url: Any): ...
    @classmethod
    def get_pool_class(cls, url: Any): ...
    def is_disconnect(self, e: Any, connection: Any, cursor: Any): ...
    def do_set_input_sizes(
        self, cursor: Any, list_of_tuples: Any, context: Any
    ) -> None: ...
    def on_connect(self): ...

dialect = PGDialect_asyncpg



import sys
from typing import Any
from typing import Callable
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Mapping
from typing import Optional
from typing import Tuple
from typing import TypeVar

from .row import Row as Row
from ..sql.base import InPlaceGenerative
from ..util import collections_abc

_TResult = TypeVar("_TResult", bound=Result)
_TScalarResult = TypeVar("_TScalarResult", bound=ScalarResult)
_TMappingResult = TypeVar("_TMappingResult", bound=MappingResult)
_TChunkedIteratorResult = TypeVar(
    "_TChunkedIteratorResult", bound=ChunkedIteratorResult
)

def tuplegetter(*indexes: Any) -> Callable[[Any], Tuple[Any, ...]]: ...

class ResultMetaData:
    @property
    def keys(self) -> RMKeyView: ...

class RMKeyView(collections_abc.KeysView[str]):
    def __init__(self, parent: Any) -> None: ...
    def __len__(self) -> int: ...
    def __iter__(self) -> Iterator[str]: ...
    def __contains__(self, item: object) -> bool: ...
    def __eq__(self, other: Any) -> bool: ...
    def __ne__(self, other: Any) -> bool: ...

class SimpleResultMetaData(ResultMetaData):
    def __init__(
        self,
        keys: Iterable[str],
        extra: Optional[Any] = ...,
        _processors: Optional[Any] = ...,
        _tuplefilter: Optional[Any] = ...,
        _translated_indexes: Optional[Any] = ...,
        _unique_filters: Optional[Any] = ...,
    ) -> None: ...

def result_tuple(fields: Any, extra: Optional[Any] = ...) -> Row: ...

class ResultInternal(InPlaceGenerative): ...

class _WithKeys:
    def keys(self) -> RMKeyView: ...

class Result(_WithKeys, ResultInternal):
    def __init__(self, cursor_metadata: ResultMetaData) -> None: ...
    def yield_per(self: _TResult, num: Any) -> _TResult: ...
    def unique(
        self: _TResult, strategy: Optional[object] = ...
    ) -> _TResult: ...
    def columns(self: _TResult, *col_expressions: object) -> _TResult: ...
    def scalars(self, index: int = ...) -> ScalarResult: ...
    def mappings(self) -> MappingResult: ...
    def __iter__(self) -> Iterator[Row]: ...
    def __next__(self) -> Row: ...
    if sys.version_info < (3, 0):
        def next(self) -> Row: ...
    def partitions(self, size: Optional[int] = ...) -> Iterator[List[Row]]: ...
    def fetchall(self) -> List[Row]: ...
    def fetchone(self) -> Optional[Row]: ...
    def fetchmany(self, size: Optional[int] = ...) -> List[Row]: ...
    def all(self) -> List[Row]: ...
    def first(self) -> Optional[Row]: ...
    def one_or_none(self) -> Optional[Row]: ...
    def scalar_one(self) -> Any: ...
    def scalar_one_or_none(self) -> Optional[Any]: ...
    def one(self) -> Row: ...
    def scalar(self) -> Optional[Any]: ...
    def freeze(self) -> FrozenResult: ...
    def merge(self, *others: Result) -> MergedResult: ...

class FilterResult(ResultInternal): ...

class ScalarResult(FilterResult):
    def __init__(self, real_result: Result, index: Any) -> None: ...
    def unique(
        self: _TScalarResult, strategy: Optional[Any] = ...
    ) -> _TScalarResult: ...
    def partitions(self, size: Optional[int] = ...) -> Iterator[List[Any]]: ...
    def fetchall(self) -> List[Any]: ...
    def fetchmany(self, size: Optional[int] = ...) -> List[Any]: ...
    def all(self) -> List[Any]: ...
    def __iter__(self) -> Iterator[Any]: ...
    def __next__(self) -> Any: ...
    if sys.version_info < (3, 0):
        def next(self) -> Any: ...
    def first(self) -> Optional[Any]: ...
    def one_or_none(self) -> Optional[Any]: ...
    def one(self) -> Any: ...

class MappingResult(_WithKeys, FilterResult):
    def __init__(self, result: Result) -> None: ...
    def unique(
        self: _TMappingResult, strategy: Optional[Any] = ...
    ) -> _TMappingResult: ...
    def columns(
        self: _TMappingResult, *col_expressions: object
    ) -> _TMappingResult: ...
    def partitions(
        self, size: Optional[int] = ...
    ) -> Iterator[List[Mapping[Any, Any]]]: ...
    def fetchall(self) -> List[Mapping[Any, Any]]: ...
    def fetchone(self) -> Mapping[Any, Any]: ...
    def fetchmany(
        self, size: Optional[int] = ...
    ) -> List[Mapping[Any, Any]]: ...
    def all(self) -> List[Mapping[Any, Any]]: ...
    def __iter__(self) -> Iterator[Mapping[Any, Any]]: ...
    def __next__(self) -> Mapping[Any, Any]: ...
    if sys.version_info < (3, 0):
        def next(self) -> Mapping[Any, Any]: ...
    def first(self) -> Optional[Mapping[Any, Any]]: ...
    def one_or_none(self) -> Optional[Mapping[Any, Any]]: ...
    def one(self) -> Mapping[Any, Any]: ...

class FrozenResult:
    metadata: Any = ...
    data: Any = ...
    def __init__(self, result: Result) -> None: ...
    def rewrite_rows(self) -> List[List[Any]]: ...
    def with_new_rows(
        self, tuple_data: List[Tuple[Any, ...]]
    ) -> FrozenResult: ...
    def __call__(self) -> IteratorResult: ...

class IteratorResult(Result):
    iterator: Any = ...
    raw: Any = ...
    def __init__(
        self,
        cursor_metadata: ResultMetaData,
        iterator: Any,
        raw: Optional[Any] = ...,
    ) -> None: ...

def null_result() -> IteratorResult: ...

class ChunkedIteratorResult(IteratorResult):
    chunks: Any = ...
    raw: Any = ...
    iterator: Any = ...
    dynamic_yield_per: Any = ...
    def __init__(
        self,
        cursor_metadata: ResultMetaData,
        chunks: Any,
        source_supports_scalars: bool = ...,
        raw: Optional[Any] = ...,
        dynamic_yield_per: bool = ...,
    ) -> None: ...
    def yield_per(
        self: _TChunkedIteratorResult, num: Any
    ) -> _TChunkedIteratorResult: ...

class MergedResult(IteratorResult):
    closed: bool = ...
    def __init__(
        self, cursor_metadata: ResultMetaData, results: Result
    ) -> None: ...
    def close(self) -> None: ...



from typing import Any
from typing import Optional

from ... import types as sqltypes
from ...sql import functions as sqlfunc

class HSTORE(sqltypes.Indexable, sqltypes.Concatenable, sqltypes.TypeEngine):
    __visit_name__: str = ...
    hashable: bool = ...
    text_type: Any = ...
    def __init__(self, text_type: Optional[Any] = ...) -> None: ...
    class Comparator(
        sqltypes.Indexable.Comparator, sqltypes.Concatenable.Comparator
    ):
        def has_key(self, other: Any): ...
        def has_all(self, other: Any): ...
        def has_any(self, other: Any): ...
        def contains(self, other: Any, **kwargs: Any): ...
        def contained_by(self, other: Any): ...
        def defined(self, key: Any): ...
        def delete(self, key: Any): ...
        def slice(self, array: Any): ...
        def keys(self): ...
        def vals(self): ...
        def array(self): ...
        def matrix(self): ...
    comparator_factory: Any = ...
    def bind_processor(self, dialect: Any): ...
    def result_processor(self, dialect: Any, coltype: Any): ...

class hstore(sqlfunc.GenericFunction):
    type: Any = ...
    name: str = ...

class _HStoreDefinedFunction(sqlfunc.GenericFunction):
    type: Any = ...
    name: str = ...

class _HStoreDeleteFunction(sqlfunc.GenericFunction):
    type: Any = ...
    name: str = ...

class _HStoreSliceFunction(sqlfunc.GenericFunction):
    type: Any = ...
    name: str = ...

class _HStoreKeysFunction(sqlfunc.GenericFunction):
    type: Any = ...
    name: str = ...

class _HStoreValsFunction(sqlfunc.GenericFunction):
    type: Any = ...
    name: str = ...

class _HStoreArrayFunction(sqlfunc.GenericFunction):
    type: Any = ...
    name: str = ...

class _HStoreMatrixFunction(sqlfunc.GenericFunction):
    type: Any = ...
    name: str = ...



from typing import Any

from . import attributes as attributes
from . import exc as exc
from .. import util as util

def populate(
    source: Any,
    source_mapper: Any,
    dest: Any,
    dest_mapper: Any,
    synchronize_pairs: Any,
    uowcommit: Any,
    flag_cascaded_pks: Any,
) -> None: ...
def bulk_populate_inherit_keys(
    source_dict: Any, source_mapper: Any, synchronize_pairs: Any
) -> None: ...
def clear(dest: Any, dest_mapper: Any, synchronize_pairs: Any) -> None: ...
def update(
    source: Any,
    source_mapper: Any,
    dest: Any,
    old_prefix: Any,
    synchronize_pairs: Any,
) -> None: ...
def populate_dict(
    source: Any, source_mapper: Any, dict_: Any, synchronize_pairs: Any
) -> None: ...
def source_modified(
    uowcommit: Any, source: Any, source_mapper: Any, synchronize_pairs: Any
): ...



from typing import Any
from typing import Optional

from .base import BINARY as BINARY
from .base import DATETIMEOFFSET as DATETIMEOFFSET
from .base import MSDialect as MSDialect
from .base import MSExecutionContext as MSExecutionContext
from .base import VARBINARY as VARBINARY
from ... import exc as exc
from ... import types as sqltypes
from ... import util as util
from ...connectors.pyodbc import PyODBCConnector as PyODBCConnector

class _ms_numeric_pyodbc:
    def bind_processor(self, dialect: Any): ...

class _MSNumeric_pyodbc(_ms_numeric_pyodbc, sqltypes.Numeric): ...
class _MSFloat_pyodbc(_ms_numeric_pyodbc, sqltypes.Float): ...

class _ms_binary_pyodbc:
    def bind_processor(self, dialect: Any): ...

class _ODBCDateTimeOffset(DATETIMEOFFSET):
    def bind_processor(self, dialect: Any): ...

class _VARBINARY_pyodbc(_ms_binary_pyodbc, VARBINARY): ...
class _BINARY_pyodbc(_ms_binary_pyodbc, BINARY): ...

class MSExecutionContext_pyodbc(MSExecutionContext):
    def pre_exec(self) -> None: ...
    def post_exec(self) -> None: ...

class MSDialect_pyodbc(PyODBCConnector, MSDialect):
    supports_sane_rowcount_returning: bool = ...
    execution_ctx_cls: Any = ...
    colspecs: Any = ...
    description_encoding: Any = ...
    use_scope_identity: Any = ...
    fast_executemany: Any = ...
    def __init__(
        self,
        description_encoding: Optional[Any] = ...,
        fast_executemany: bool = ...,
        **params: Any,
    ) -> None: ...
    def on_connect(self): ...
    def do_executemany(
        self,
        cursor: Any,
        statement: Any,
        parameters: Any,
        context: Optional[Any] = ...,
    ) -> None: ...
    def is_disconnect(self, e: Any, connection: Any, cursor: Any): ...

dialect = MSDialect_pyodbc



from typing import Any
from typing import Collection
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple
from typing import Type
from typing import TypeVar

from .base import Connectable
from ..sql.schema import Table

_TInspector = TypeVar("_TInspector", bound=Inspector)

def cache(fn: Any, self: Any, con: Any, *args: Any, **kw: Any) -> Any: ...

class Inspector:
    def __init__(self, bind: Connectable): ...
    @classmethod
    def from_engine(
        cls: Type[_TInspector], bind: Connectable
    ) -> _TInspector: ...
    @property
    def default_schema_name(self) -> str: ...
    def get_schema_names(self) -> List[str]: ...
    def get_table_names(self, schema: Optional[str] = ...) -> List[str]: ...
    def has_table(
        self, table_name: str, schema: Optional[str] = ...
    ) -> bool: ...
    def has_sequence(
        self, sequence_name: str, schema: Optional[str] = ...
    ) -> bool: ...
    def get_sorted_table_and_fkc_names(
        self, schema: Optional[str] = ...
    ) -> List[Tuple[Optional[str], List[Tuple[str, str]]]]: ...
    def get_temp_table_names(self) -> List[str]: ...
    def get_temp_view_names(self) -> List[str]: ...
    def get_table_options(
        self, table_name: str, schema: Optional[str] = ..., **kw: Any
    ) -> Dict[str, Any]: ...
    def get_view_names(self, schema: Optional[str] = ...) -> List[str]: ...
    def get_sequence_names(self, schema: Optional[str] = ...) -> List[str]: ...
    def get_view_definition(
        self, view_name: str, schema: Optional[str] = ...
    ) -> Any: ...
    def get_columns(
        self, table_name: str, schema: Optional[str] = ..., **kw: Any
    ) -> List[Dict[str, Any]]: ...
    def get_pk_constraint(
        self, table_name: str, schema: Optional[str] = ..., **kw: Any
    ) -> Dict[str, Any]: ...
    def get_foreign_keys(
        self, table_name: str, schema: Optional[str] = ..., **kw: Any
    ) -> List[Dict[str, Any]]: ...
    def get_indexes(
        self, table_name: str, schema: Optional[str] = ..., **kw: Any
    ) -> List[Dict[str, Any]]: ...
    def get_unique_constraints(
        self, table_name: str, schema: Optional[str] = ..., **kw: Any
    ) -> List[Dict[str, Any]]: ...
    def get_table_comment(
        self, table_name: str, schema: Optional[str] = ..., **kw: Any
    ) -> Dict[str, Any]: ...
    def get_check_constraints(
        self, table_name: str, schema: Optional[str] = ..., **kw: Any
    ) -> List[Dict[str, Any]]: ...
    def reflecttable(
        self,
        table: Table,
        include_columns: Collection[str],
        exclude_columns: Collection[str] = ...,
        resolve_fks: bool = ...,
        _extend_on: Set[Table] = ...,
    ) -> None: ...
    def reflect_table(
        self,
        table: Table,
        include_columns: Collection[str],
        exclude_columns: Collection[str] = ...,
        resolve_fks: bool = ...,
        _extend_on: Set[Table] = ...,
    ) -> None: ...



# fmt: off
from typing import Any
from typing import Optional

from sqlalchemy.cprocessors import DecimalResultProcessor as DecimalResultProcessor
from sqlalchemy.cprocessors import int_to_boolean as int_to_boolean
from sqlalchemy.cprocessors import str_to_date as str_to_date
from sqlalchemy.cprocessors import str_to_datetime as str_to_datetime
from sqlalchemy.cprocessors import str_to_time as str_to_time
from sqlalchemy.cprocessors import to_float as to_float
from sqlalchemy.cprocessors import to_str as to_str
from sqlalchemy.cprocessors import UnicodeResultProcessor as UnicodeResultProcessor
from . import util as util
# fmt: on

def str_to_datetime_processor_factory(regexp: Any, type_: Any): ...
def py_fallback(): ...
def to_unicode_processor_factory(
    encoding: Any, errors: Optional[Any] = ...
): ...
def to_conditional_unicode_processor_factory(
    encoding: Any, errors: Optional[Any] = ...
): ...
def to_decimal_processor_factory(target_class: Any, scale: Any): ...



from typing import Any

from .. import event

class DDLEvents(event.Events):
    def before_create(
        self, target: Any, connection: Any, **kw: Any
    ) -> None: ...
    def after_create(
        self, target: Any, connection: Any, **kw: Any
    ) -> None: ...
    def before_drop(self, target: Any, connection: Any, **kw: Any) -> None: ...
    def after_drop(self, target: Any, connection: Any, **kw: Any) -> None: ...
    def before_parent_attach(self, target: Any, parent: Any) -> None: ...
    def after_parent_attach(self, target: Any, parent: Any) -> None: ...
    def column_reflect(
        self, inspector: Any, table: Any, column_info: Any
    ) -> None: ...



from typing import Any

from . import legacy as legacy
from . import registry as registry
from .. import exc as exc
from .. import util as util
from ..util import threading as threading
from ..util.concurrency import AsyncAdaptedLock as AsyncAdaptedLock

class RefCollection(util.MemoizedSlots): ...

class _empty_collection:
    def append(self, element: Any) -> None: ...
    def extend(self, other: Any) -> None: ...
    def remove(self, element: Any) -> None: ...
    def __iter__(self) -> Any: ...
    def clear(self) -> None: ...

class _ClsLevelDispatch(RefCollection):
    name: Any = ...
    clsname: Any = ...
    arg_names: Any = ...
    has_kw: Any = ...
    legacy_signatures: Any = ...
    def __init__(self, parent_dispatch_cls: Any, fn: Any): ...
    def insert(self, event_key: Any, propagate: Any) -> None: ...
    def append(self, event_key: Any, propagate: Any) -> None: ...
    def update_subclass(self, target: Any) -> None: ...
    def remove(self, event_key: Any) -> None: ...
    def clear(self) -> None: ...
    def for_modify(self, obj: Any): ...

class _InstanceLevelDispatch(RefCollection): ...

class _EmptyListener(_InstanceLevelDispatch):
    propagate: Any = ...
    listeners: Any = ...
    parent: Any = ...
    parent_listeners: Any = ...
    name: Any = ...
    def __init__(self, parent: Any, target_cls: Any) -> None: ...
    def for_modify(self, obj: Any): ...
    exec_once: Any = ...
    exec_once_unless_exception: Any = ...
    insert: Any = ...
    append: Any = ...
    remove: Any = ...
    clear: Any = ...
    def __call__(self, *args: Any, **kw: Any) -> None: ...
    def __len__(self): ...
    def __iter__(self) -> Any: ...
    def __bool__(self): ...
    __nonzero__: Any = ...

class _CompoundListener(_InstanceLevelDispatch):
    def exec_once(self, *args: Any, **kw: Any) -> None: ...
    def exec_once_unless_exception(self, *args: Any, **kw: Any) -> None: ...
    def __call__(self, *args: Any, **kw: Any) -> None: ...
    def __len__(self): ...
    def __iter__(self) -> Any: ...
    def __bool__(self): ...
    __nonzero__: Any = ...

class _ListenerCollection(_CompoundListener):
    parent_listeners: Any = ...
    parent: Any = ...
    name: Any = ...
    listeners: Any = ...
    propagate: Any = ...
    def __init__(self, parent: Any, target_cls: Any) -> None: ...
    def for_modify(self, obj: Any): ...
    def insert(self, event_key: Any, propagate: Any) -> None: ...
    def append(self, event_key: Any, propagate: Any) -> None: ...
    def remove(self, event_key: Any) -> None: ...
    def clear(self) -> None: ...

class _JoinedListener(_CompoundListener):
    parent: Any = ...
    name: Any = ...
    local: Any = ...
    parent_listeners: Any = ...
    def __init__(self, parent: Any, name: Any, local: Any) -> None: ...
    @property
    def listeners(self): ...
    def for_modify(self, obj: Any): ...
    def insert(self, event_key: Any, propagate: Any) -> None: ...
    def append(self, event_key: Any, propagate: Any) -> None: ...
    def remove(self, event_key: Any) -> None: ...
    def clear(self) -> None: ...



from typing import Any
from typing import Optional

from .. import inspect as inspect
from .. import util as util
from ..sql import and_ as and_
from ..sql import operators as operators

class UnevaluatableError(Exception): ...

class _NoObject(operators.ColumnOperators):
    def operate(self, *arg: Any, **kw: Any) -> None: ...
    def reverse_operate(self, *arg: Any, **kw: Any) -> None: ...

class EvaluatorCompiler:
    target_cls: Any = ...
    def __init__(self, target_cls: Optional[Any] = ...) -> None: ...
    def process(self, *clauses: Any): ...
    def visit_grouping(self, clause: Any): ...
    def visit_null(self, clause: Any): ...
    def visit_false(self, clause: Any): ...
    def visit_true(self, clause: Any): ...
    def visit_column(self, clause: Any): ...
    def visit_tuple(self, clause: Any): ...
    def visit_clauselist(self, clause: Any): ...
    def visit_binary(self, clause: Any): ...
    def visit_unary(self, clause: Any): ...
    def visit_bindparam(self, clause: Any): ...



from types import ModuleType
from typing import Any
from typing import Callable
from typing import Dict
from typing import Generic
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Mapping
from typing import NoReturn
from typing import Optional
from typing import overload
from typing import Tuple
from typing import Type
from typing import TypeVar
from typing import Union

from . import roles
from .elements import ColumnElement
from .traversals import HasCacheKey
from .traversals import HasCopyInternals
from .visitors import ClauseVisitor
from .. import util
from ..engine import Connection
from ..engine import Engine
from ..util import HasMemoized as HasMemoized
from ..util import langhelpers

_T = TypeVar("_T")
_SC = TypeVar("_SC", bound=SingletonConstant)
_O = TypeVar("_O", bound=Options)
_E = TypeVar("_E", bound=Executable)
_CE = TypeVar("_CE", bound=ColumnElement[Any])
_OCE = TypeVar("_OCE", bound=ColumnElement[Any])

coercions: ModuleType
elements: ModuleType
type_api: ModuleType
PARSE_AUTOCOMMIT: langhelpers._symbol
NO_ARG: langhelpers._symbol

class Immutable:
    def unique_params(self, *optionaldict: Any, **kwargs: Any) -> NoReturn: ...
    def params(self, *optionaldict: Any, **kwargs: Any) -> NoReturn: ...

class SingletonConstant(Immutable):
    def __new__(cls: Type[_SC], *arg: Any, **kw: Any) -> _SC: ...

class _DialectArgView(util.collections_abc.MutableMapping[str, Any]):
    obj: Any = ...
    def __init__(self, obj: Any) -> None: ...
    def __getitem__(self, key: str) -> Any: ...
    def __setitem__(self, key: str, value: Any) -> None: ...
    def __delitem__(self, key: str) -> None: ...
    def __len__(self) -> int: ...
    def __iter__(self) -> Iterator[str]: ...

class _DialectArgDict(util.collections_abc.MutableMapping[str, Any]):
    def __init__(self) -> None: ...
    def __len__(self) -> int: ...
    def __iter__(self) -> Iterator[str]: ...
    def __getitem__(self, key: str) -> Any: ...
    def __setitem__(self, key: str, value: Any) -> None: ...
    def __delitem__(self, key: str) -> None: ...

class DialectKWArgs:
    @classmethod
    def argument_for(
        cls, dialect_name: Any, argument_name: Any, default: Any
    ) -> None: ...
    @util.memoized_property
    def dialect_kwargs(self) -> _DialectArgView: ...
    @property
    def kwargs(self) -> _DialectArgView: ...
    @util.memoized_property
    def dialect_options(self) -> util.PopulateDict[str, _DialectArgDict]: ...

class CompileState:
    plugins: Any = ...
    @classmethod
    def create_for_statement(
        cls, statement: Any, compiler: Any, **kw: Any
    ) -> Any: ...
    statement: Any = ...
    def __init__(self, statement: Any, compiler: Any, **kw: Any) -> None: ...
    @classmethod
    def get_plugin_class(cls, statement: Any) -> Optional[Any]: ...
    @classmethod
    def plugin_for(
        cls, plugin_name: Any, visit_name: Any
    ) -> Callable[[_T], _T]: ...

class Generative(HasMemoized): ...
class InPlaceGenerative(HasMemoized): ...
class HasCompileState(Generative): ...

class _MetaOptions(type):
    def __init__(cls, classname: Any, bases: Any, dict_: Any) -> None: ...
    def __add__(self: Type[_T], other: Any) -> _T: ...

class Options(metaclass=_MetaOptions):
    def __init__(self, **kw: Any) -> None: ...
    def __add__(self: _O, other: Any) -> _O: ...
    def __eq__(self, other: Any) -> bool: ...
    @classmethod
    def isinstance(cls, klass: Any) -> bool: ...
    def add_to_element(self: _O, name: Any, value: Any) -> _O: ...
    @classmethod
    def safe_merge(cls: Type[_O], other: Any) -> _O: ...
    @classmethod
    def from_execution_options(
        cls,
        key: Any,
        attrs: Any,
        exec_options: Any,
        statement_exec_options: Any,
    ) -> Tuple[Any, Any]: ...

class CacheableOptions(Options, HasCacheKey): ...

class ExecutableOption(HasCopyInternals, HasCacheKey):
    __visit_name__: str = ...

class Executable(roles.CoerceTextStatementRole, Generative):
    supports_execution: bool = ...
    is_select: bool = ...
    is_update: bool = ...
    is_insert: bool = ...
    is_text: bool = ...
    is_delete: bool = ...
    is_dml: bool = ...
    def options(self: _E, *options: Any) -> _E: ...
    def execution_options(self: _E, **kw: Any) -> _E: ...
    def get_execution_options(self) -> Any: ...
    def execute(self, *multiparams: Any, **params: Any) -> Any: ...
    def scalar(self, *multiparams: Any, **params: Any) -> Any: ...
    @property
    def bind(self) -> Optional[Union[Engine, Connection]]: ...

class prefix_anon_map(Dict[str, str]):
    def __missing__(self, key: str) -> str: ...

class SchemaEventTarget: ...

class SchemaVisitor(ClauseVisitor):
    __traverse_options__: Any = ...

class ColumnCollection(Generic[_CE]):
    def __init__(
        self, columns: Optional[Iterable[Tuple[str, _CE]]] = ...
    ) -> None: ...
    def keys(self) -> List[str]: ...
    def values(self) -> List[_CE]: ...
    def items(self) -> List[Tuple[str, _CE]]: ...
    def __bool__(self) -> bool: ...
    def __len__(self) -> int: ...
    def __iter__(self) -> Iterator[_CE]: ...
    def __getitem__(self, key: str) -> _CE: ...
    def __getattr__(self, key: str) -> _CE: ...
    def __contains__(self, key: str) -> bool: ...
    def compare(self, other: Any) -> bool: ...
    def __eq__(self, other: Any) -> bool: ...
    @overload
    def get(self, key: str) -> Optional[_CE]: ...
    @overload
    def get(self, key: str, default: _T) -> Union[_CE, _T]: ...
    def __setitem__(self, key: str, value: _CE) -> None: ...
    def __delitem__(self, key: str) -> None: ...
    def __setattr__(self, key: str, obj: _CE) -> None: ...
    def clear(self) -> None: ...
    def remove(self, column: _CE) -> None: ...
    def update(
        self, iter_: Union[Mapping[str, _CE], Iterable[Tuple[str, _CE]]]
    ) -> None: ...
    __hash__: Any = ...
    def add(self, column: _CE, key: Optional[str] = ...) -> None: ...
    def contains_column(self, col: ColumnElement[Any]) -> bool: ...
    def as_immutable(self) -> ImmutableColumnCollection[_CE]: ...
    def corresponding_column(
        self, column: ColumnElement[Any], require_embedded: bool = ...
    ) -> Optional[_CE]: ...

class DedupeColumnCollection(ColumnCollection[_CE]):
    def add(self, column: _CE, key: Optional[str] = ...) -> None: ...
    def extend(self, iter_: Iterable[_CE]) -> None: ...
    def remove(self, column: _CE) -> None: ...
    def replace(self, column: _CE) -> None: ...

class ImmutableColumnCollection(
    util.ImmutableContainer, ColumnCollection[_CE]
):
    def __init__(self, collection: ColumnCollection[_CE]) -> None: ...
    def add(self, column: _CE, key: Optional[str] = ...) -> NoReturn: ...
    def extend(self, iter_: Iterable[_CE]) -> NoReturn: ...
    def remove(self, column: _CE) -> NoReturn: ...

class ColumnSet(util.ordered_column_set[_CE]):
    def contains_column(self, col: ColumnElement[Any]) -> bool: ...
    def extend(self, cols: Iterable[_CE]) -> None: ...
    def __add__(self, other: Iterable[_OCE]) -> List[Union[_CE, _OCE]]: ...  # type: ignore[override]
    def __eq__(self, other: Any) -> Any: ...
    def __hash__(self) -> int: ...  # type: ignore[override]



from typing import Any

from .base import DATE as DATE
from .base import DATETIME as DATETIME
from .base import SQLiteDialect as SQLiteDialect
from ... import exc as exc
from ... import pool as pool
from ... import util as util

class _SQLite_pysqliteTimeStamp(DATETIME):
    def bind_processor(self, dialect: Any): ...
    def result_processor(self, dialect: Any, coltype: Any): ...

class _SQLite_pysqliteDate(DATE):
    def bind_processor(self, dialect: Any): ...
    def result_processor(self, dialect: Any, coltype: Any): ...

class SQLiteDialect_pysqlite(SQLiteDialect):
    default_paramstyle: str = ...
    colspecs: Any = ...
    description_encoding: Any = ...
    driver: str = ...
    @classmethod
    def dbapi(cls): ...
    @classmethod
    def get_pool_class(cls, url: Any): ...
    def set_isolation_level(self, connection: Any, level: Any): ...
    def on_connect(self): ...
    def create_connect_args(self, url: Any): ...
    def is_disconnect(self, e: Any, connection: Any, cursor: Any): ...

dialect = SQLiteDialect_pysqlite



from typing import Any

from .base import BIGINT as BIGINT
from .base import BINARY as BINARY
from .base import BIT as BIT
from .base import CHAR as CHAR
from .base import DATE as DATE
from .base import DATETIME as DATETIME
from .base import FLOAT as FLOAT
from .base import IMAGE as IMAGE
from .base import INT as INT
from .base import INTEGER as INTEGER
from .base import MONEY as MONEY
from .base import NCHAR as NCHAR
from .base import NUMERIC as NUMERIC
from .base import NVARCHAR as NVARCHAR
from .base import SMALLINT as SMALLINT
from .base import SMALLMONEY as SMALLMONEY
from .base import TEXT as TEXT
from .base import TIME as TIME
from .base import TINYINT as TINYINT
from .base import UNICHAR as UNICHAR
from .base import UNITEXT as UNITEXT
from .base import UNIVARCHAR as UNIVARCHAR
from .base import VARBINARY as VARBINARY
from .base import VARCHAR as VARCHAR

dialect: Any



from typing import Any
from typing import Generic
from typing import Optional
from typing import overload
from typing import Type
from typing import TypeVar
from typing import Union

from typing_extensions import Protocol

from . import sqltypes
from . import type_api
from .base import ColumnCollection
from .base import Executable
from .base import Generative
from .base import HasMemoized
from .elements import AsBoolean
from .elements import BinaryExpression
from .elements import ClauseElement
from .elements import ClauseList
from .elements import ColumnElement
from .elements import FunctionFilter
from .elements import Grouping
from .elements import NamedColumn
from .elements import Over
from .elements import TableValuedColumn
from .elements import WithinGroup
from .schema import Sequence
from .selectable import FromClause
from .selectable import Select
from .selectable import TableValuedAlias
from .visitors import TraversibleType

_T_co = TypeVar("_T_co", covariant=True)
_TE = TypeVar("_TE", bound=type_api.TypeEngine[Any])
_FE = TypeVar("_FE", bound=FunctionElement[Any])

def register_function(
    identifier: str, fn: Any, package: str = ...
) -> None: ...

class FunctionElement(  # type: ignore[misc]
    Executable, ColumnElement[_TE], FromClause, Generative, Generic[_TE]
):
    packagenames: Any = ...
    clause_expr: Any = ...
    def __init__(self, *clauses: Any, **kwargs: Any) -> None: ...
    @overload
    def scalar_table_valued(
        self, name: Any, type_: None = ...
    ) -> ScalarFunctionColumn[sqltypes.NullType]: ...
    @overload
    def scalar_table_valued(
        self, name: Any, type_: Union[_TE, Type[_TE]]
    ) -> ScalarFunctionColumn[_TE]: ...
    def table_valued(self, *expr: Any, **kw: Any) -> TableValuedAlias: ...  # type: ignore[override]
    def column_valued(
        self, name: Optional[Any] = ...
    ) -> TableValuedColumn[Any]: ...
    @property
    def columns(self) -> ColumnCollection[ColumnElement[Any]]: ...  # type: ignore[override]
    @HasMemoized.memoized_attribute
    def clauses(self) -> ClauseList[Any]: ...
    def over(
        self,
        partition_by: Optional[ClauseElement] = ...,
        order_by: Optional[ClauseElement] = ...,
        rows: Optional[Any] = ...,
        range_: Optional[Any] = ...,
    ) -> Over[_TE]: ...
    def within_group(self, *order_by: Any) -> WithinGroup[_TE]: ...
    @overload
    def filter(self: _FE) -> _FE: ...
    @overload
    def filter(
        self, __criteria: Any, *criterion: Any
    ) -> FunctionFilter[_TE]: ...
    def as_comparison(
        self, left_index: int, right_index: int
    ) -> FunctionAsBinary: ...
    def within_group_type(
        self, within_group: Any
    ) -> Optional[type_api.TypeEngine[Any]]: ...
    def alias(self, name: Optional[Any] = ...) -> TableValuedAlias: ...  # type: ignore[override]
    def select(self) -> Select: ...  # type: ignore[override]
    def scalar(self) -> Any: ...  # type: ignore[override]
    def execute(self) -> Any: ...  # type: ignore[override]
    def self_group(
        self: _FE, against: Optional[Any] = ...
    ) -> Union[_FE, Grouping[_TE], AsBoolean[_FE]]: ...

class FunctionAsBinary(BinaryExpression[sqltypes.Boolean]):
    sql_function: FunctionElement[Any] = ...
    left_index: int = ...
    right_index: int = ...
    operator: Any = ...
    type: sqltypes.Boolean = ...
    negate: Any = ...
    modifiers: Any = ...
    left: ClauseElement = ...
    right: ClauseElement = ...
    def __init__(
        self, fn: FunctionElement[Any], left_index: int, right_index: int
    ) -> None: ...

class ScalarFunctionColumn(NamedColumn[_TE]):
    __visit_name__: str = ...
    is_literal: bool = ...
    table: Any = ...
    fn: ClauseElement = ...
    name: Any = ...
    type: _TE = ...  # type: ignore[assignment]
    @overload
    def __init__(
        self: ScalarFunctionColumn[sqltypes.NullType],
        fn: ClauseElement,
        name: str,
        type_: None = ...,
    ) -> None: ...
    @overload
    def __init__(
        self, fn: ClauseElement, name: str, type_: Union[_TE, Type[_TE]]
    ) -> None: ...

class _FunctionGenerator:
    opts: Any = ...
    def __init__(self, **opts: Any) -> None: ...
    def __getattr__(self, name: str) -> _FunctionGenerator: ...
    @overload
    def __call__(
        self, *c: Any, type_: None = ..., **kwargs: Any
    ) -> Function[sqltypes.NullType]: ...
    @overload
    def __call__(
        self, *c: Any, type_: Union[_TE, Type[_TE]], **kwargs: Any
    ) -> Function[_TE]: ...

func: _FunctionGenerator
modifier: _FunctionGenerator

class _TypeDescriptor(Protocol[_T_co]):
    @overload
    def __get__(self, instance: None, owner: Any) -> _T_co: ...
    @overload
    def __get__(self, instance: GenericFunction[_TE], owner: Any) -> _TE: ...

class Function(FunctionElement[_TE]):
    __visit_name__: str = ...
    type: _TypeDescriptor[sqltypes.NullType] = ...  # type: ignore[assignment]
    packagenames: Any = ...
    name: Any = ...
    @overload
    def __init__(
        self: Function[sqltypes.NullType],
        name: Any,
        *clauses: Any,
        type_: None = ...,
        **kw: Any,
    ) -> None: ...
    @overload
    def __init__(
        self, name: Any, *clauses: Any, type_: Union[_TE, Type[_TE]], **kw: Any
    ) -> None: ...

class _GenericMeta(TraversibleType):
    def __init__(cls, clsname: Any, bases: Any, clsdict: Any) -> None: ...

class GenericFunction(Function[_TE], metaclass=_GenericMeta):
    coerce_arguments: bool = ...
    inherit_cache: bool = ...
    packagenames: Any = ...
    clause_expr: Any = ...
    type: _TypeDescriptor[sqltypes.NullType] = ...
    @overload
    def __init__(
        self: GenericFunction[sqltypes.NullType],
        *args: Any,
        type_: None = ...,
        **kwargs: Any,
    ) -> None: ...
    @overload
    def __init__(
        self, *args: Any, type_: Union[_TE, Type[_TE]], **kwargs: Any
    ) -> None: ...

class next_value(GenericFunction[_TE]):
    type: _TypeDescriptor[sqltypes.Integer] = ...  # type: ignore[assignment]
    name: str = ...
    sequence: Sequence[_TE] = ...
    def __init__(self, seq: Sequence[_TE], **kw: Any) -> None: ...
    def compare(self, other: Any, **kw: Any) -> bool: ...

class AnsiFunction(GenericFunction[_TE]):
    inherit_cache: bool = ...

class ReturnTypeFromArgs(GenericFunction[_TE]):
    inherit_cache: bool = ...
    def __init__(self, *args: Any, **kwargs: Any) -> None: ...

class coalesce(ReturnTypeFromArgs[_TE]):
    inherit_cache: bool = ...

class max(ReturnTypeFromArgs[_TE]):
    inherit_cache: bool = ...

class min(ReturnTypeFromArgs[_TE]):
    inherit_cache: bool = ...

class sum(ReturnTypeFromArgs[_TE]):
    inherit_cache: bool = ...

class now(GenericFunction[_TE]):
    type: _TypeDescriptor[Type[sqltypes.DateTime]] = ...  # type: ignore[assignment]
    inherit_cache: bool = ...
    @overload
    def __init__(
        self: now[sqltypes.DateTime],
        *args: Any,
        type_: None = ...,
        **kwargs: Any,
    ) -> None: ...
    @overload
    def __init__(
        self, *args: Any, type_: Union[_TE, Type[_TE]], **kwargs: Any
    ) -> None: ...

class concat(GenericFunction[_TE]):
    type: _TypeDescriptor[Type[sqltypes.String]] = ...  # type: ignore[assignment]
    inherit_cache: bool = ...
    @overload
    def __init__(
        self: concat[sqltypes.String],
        *args: Any,
        type_: None = ...,
        **kwargs: Any,
    ) -> None: ...
    @overload
    def __init__(
        self, *args: Any, type_: Union[_TE, Type[_TE]], **kwargs: Any
    ) -> None: ...

class char_length(GenericFunction[_TE]):
    type: _TypeDescriptor[Type[sqltypes.Integer]] = ...  # type: ignore[assignment]
    inherit_cache: bool = ...
    @overload
    def __init__(
        self: char_length[sqltypes.Integer],
        arg: Any,
        type_: None = ...,
        **kwargs: Any,
    ) -> None: ...
    @overload
    def __init__(
        self, arg: Any, type_: Union[_TE, Type[_TE]], **kwargs: Any
    ) -> None: ...

class random(GenericFunction[_TE]):
    inherit_cache: bool = ...

class count(GenericFunction[_TE]):
    type: _TypeDescriptor[Type[sqltypes.Integer]] = ...  # type: ignore[assignment]
    inherit_cache: bool = ...
    @overload
    def __init__(
        self: count[sqltypes.Integer],
        expression: Optional[Any] = ...,
        *,
        type_: None = ...,
        **kwargs: Any,
    ) -> None: ...
    @overload
    def __init__(
        self,
        expression: Optional[Any] = ...,
        *,
        type_: Union[_TE, Type[_TE]],
        **kwargs: Any,
    ) -> None: ...

class current_date(AnsiFunction[_TE]):
    type: _TypeDescriptor[Type[sqltypes.Date]] = ...  # type: ignore[assignment]
    inherit_cache: bool = ...
    @overload
    def __init__(
        self: current_date[sqltypes.Date],
        *args: Any,
        type_: None = ...,
        **kwargs: Any,
    ) -> None: ...
    @overload
    def __init__(
        self, *args: Any, type_: Union[_TE, Type[_TE]], **kwargs: Any
    ) -> None: ...

class current_time(AnsiFunction[_TE]):
    type: _TypeDescriptor[Type[sqltypes.Time]] = ...  # type: ignore[assignment]
    inherit_cache: bool = ...
    @overload
    def __init__(
        self: current_time[sqltypes.Time],
        *args: Any,
        type_: None = ...,
        **kwargs: Any,
    ) -> None: ...
    @overload
    def __init__(
        self, *args: Any, type_: Union[_TE, Type[_TE]], **kwargs: Any
    ) -> None: ...

class current_timestamp(AnsiFunction[_TE]):
    type: _TypeDescriptor[Type[sqltypes.DateTime]] = ...  # type: ignore[assignment]
    inherit_cache: bool = ...
    @overload
    def __init__(
        self: current_timestamp[sqltypes.DateTime],
        *args: Any,
        type_: None = ...,
        **kwargs: Any,
    ) -> None: ...
    @overload
    def __init__(
        self, *args: Any, type_: Union[_TE, Type[_TE]], **kwargs: Any
    ) -> None: ...

class current_user(AnsiFunction[_TE]):
    type: _TypeDescriptor[Type[sqltypes.String]] = ...  # type: ignore[assignment]
    inherit_cache: bool = ...
    @overload
    def __init__(
        self: current_user[sqltypes.String],
        *args: Any,
        type_: None = ...,
        **kwargs: Any,
    ) -> None: ...
    @overload
    def __init__(
        self, *args: Any, type_: Union[_TE, Type[_TE]], **kwargs: Any
    ) -> None: ...

class localtime(AnsiFunction[_TE]):
    type: _TypeDescriptor[Type[sqltypes.DateTime]] = ...  # type: ignore[assignment]
    inherit_cache: bool = ...
    @overload
    def __init__(
        self: localtime[sqltypes.DateTime],
        *args: Any,
        type_: None = ...,
        **kwargs: Any,
    ) -> None: ...
    @overload
    def __init__(
        self, *args: Any, type_: Union[_TE, Type[_TE]], **kwargs: Any
    ) -> None: ...

class localtimestamp(AnsiFunction[_TE]):
    type: _TypeDescriptor[Type[sqltypes.DateTime]] = ...  # type: ignore[assignment]
    inherit_cache: bool = ...
    @overload
    def __init__(
        self: localtimestamp[sqltypes.DateTime],
        *args: Any,
        type_: None = ...,
        **kwargs: Any,
    ) -> None: ...
    @overload
    def __init__(
        self, *args: Any, type_: Union[_TE, Type[_TE]], **kwargs: Any
    ) -> None: ...

class session_user(AnsiFunction[_TE]):
    type: _TypeDescriptor[Type[sqltypes.String]] = ...  # type: ignore[assignment]
    inherit_cache: bool = ...
    @overload
    def __init__(
        self: session_user[sqltypes.String],
        *args: Any,
        type_: None = ...,
        **kwargs: Any,
    ) -> None: ...
    @overload
    def __init__(
        self, *args: Any, type_: Union[_TE, Type[_TE]], **kwargs: Any
    ) -> None: ...

class sysdate(AnsiFunction[_TE]):
    type: _TypeDescriptor[Type[sqltypes.DateTime]] = ...  # type: ignore[assignment]
    inherit_cache: bool = ...
    @overload
    def __init__(
        self: sysdate[sqltypes.DateTime],
        *args: Any,
        type_: None = ...,
        **kwargs: Any,
    ) -> None: ...
    @overload
    def __init__(
        self, *args: Any, type_: Union[_TE, Type[_TE]], **kwargs: Any
    ) -> None: ...

class user(AnsiFunction[_TE]):
    type: _TypeDescriptor[Type[sqltypes.String]] = ...  # type: ignore[assignment]
    inherit_cache: bool = ...
    @overload
    def __init__(
        self: user[sqltypes.String],
        *args: Any,
        type_: None = ...,
        **kwargs: Any,
    ) -> None: ...
    @overload
    def __init__(
        self, *args: Any, type_: Union[_TE, Type[_TE]], **kwargs: Any
    ) -> None: ...

class array_agg(GenericFunction[_TE]):
    type: _TypeDescriptor[Type[sqltypes.ARRAY[Any]]] = ...  # type: ignore[assignment]
    inherit_cache: bool = ...
    @overload
    def __init__(
        self: array_agg[sqltypes.ARRAY[Any]],
        *args: Any,
        type_: None = ...,
        **kwargs: Any,
    ) -> None: ...
    @overload
    def __init__(
        self, *args: Any, type_: Union[_TE, Type[_TE]], **kwargs: Any
    ) -> None: ...

class OrderedSetAgg(GenericFunction[_TE]):
    array_for_multi_clause: bool = ...
    inherit_cache: bool = ...
    def within_group_type(
        self, within_group: Any
    ) -> type_api.TypeEngine[Any]: ...

class mode(OrderedSetAgg[_TE]):
    inherit_cache: bool = ...

class percentile_cont(OrderedSetAgg[_TE]):
    array_for_multi_clause: bool = ...
    inherit_cache: bool = ...

class percentile_disc(OrderedSetAgg[_TE]):
    array_for_multi_clause: bool = ...
    inherit_cache: bool = ...

class rank(GenericFunction[_TE]):
    type: _TypeDescriptor[sqltypes.Integer] = ...  # type: ignore[assignment]
    inherit_cache: bool = ...
    @overload
    def __init__(
        self: rank[sqltypes.Integer],
        *args: Any,
        type_: None = ...,
        **kwargs: Any,
    ) -> None: ...
    @overload
    def __init__(
        self, *args: Any, type_: Union[_TE, Type[_TE]], **kwargs: Any
    ) -> None: ...

class dense_rank(GenericFunction[_TE]):
    type: _TypeDescriptor[sqltypes.Integer] = ...  # type: ignore[assignment]
    inherit_cache: bool = ...
    @overload
    def __init__(
        self: dense_rank[sqltypes.Integer],
        *args: Any,
        type_: None = ...,
        **kwargs: Any,
    ) -> None: ...
    @overload
    def __init__(
        self, *args: Any, type_: Union[_TE, Type[_TE]], **kwargs: Any
    ) -> None: ...

class percent_rank(GenericFunction[_TE]):
    type: _TypeDescriptor[sqltypes.Numeric] = ...  # type: ignore[assignment]
    inherit_cache: bool = ...
    @overload
    def __init__(
        self: percent_rank[sqltypes.Numeric],
        *args: Any,
        type_: None = ...,
        **kwargs: Any,
    ) -> None: ...
    @overload
    def __init__(
        self, *args: Any, type_: Union[_TE, Type[_TE]], **kwargs: Any
    ) -> None: ...

class cume_dist(GenericFunction[_TE]):
    type: _TypeDescriptor[sqltypes.Numeric] = ...  # type: ignore[assignment]
    inherit_cache: bool = ...
    @overload
    def __init__(
        self: cume_dist[sqltypes.Numeric],
        *args: Any,
        type_: None = ...,
        **kwargs: Any,
    ) -> None: ...
    @overload
    def __init__(
        self, *args: Any, type_: Union[_TE, Type[_TE]], **kwargs: Any
    ) -> None: ...

class cube(GenericFunction[_TE]):
    inherit_cache: bool = ...

class rollup(GenericFunction[_TE]):
    inherit_cache: bool = ...

class grouping_sets(GenericFunction[_TE]):
    inherit_cache: bool = ...



from typing import Any

from .. import exc as exc
from .. import util as util
from ..sql import sqltypes as sqltypes

def compiles(class_: Any, *specs: Any): ...
def deregister(class_: Any) -> None: ...

class _dispatcher:
    specs: Any = ...
    def __init__(self) -> None: ...
    def __call__(self, element: Any, compiler: Any, **kw: Any): ...



from typing import Any

from .base import PGDialect as PGDialect
from .base import PGExecutionContext as PGExecutionContext
from ... import processors as processors
from ... import types as sqltypes
from ... import util as util

class PGNumeric(sqltypes.Numeric):
    def bind_processor(self, dialect: Any): ...
    def result_processor(self, dialect: Any, coltype: Any): ...

class PGExecutionContext_pypostgresql(PGExecutionContext): ...

class PGDialect_pypostgresql(PGDialect):
    driver: str = ...
    supports_unicode_statements: bool = ...
    supports_unicode_binds: bool = ...
    description_encoding: Any = ...
    default_paramstyle: str = ...
    supports_sane_rowcount: bool = ...
    supports_sane_multi_rowcount: bool = ...
    execution_ctx_cls: Any = ...
    colspecs: Any = ...
    @classmethod
    def dbapi(cls): ...
    def dbapi_exception_translation_map(self): ...
    def create_connect_args(self, url: Any): ...
    def is_disconnect(self, e: Any, connection: Any, cursor: Any): ...

dialect = PGDialect_pypostgresql



# fmt: off
from typing import Any
from typing import Optional

from .array import ARRAY as ARRAY
from ... import util as util
from ...sql import coercions as coercions
from ...sql import elements as elements
from ...sql import expression as expression
from ...sql import functions as functions
from ...sql import roles as roles
from ...sql import schema as schema
from ...sql.schema import ColumnCollectionConstraint as ColumnCollectionConstraint
# fmt: on

class aggregate_order_by(expression.ColumnElement):
    __visit_name__: str = ...
    stringify_dialect: str = ...
    target: Any = ...
    order_by: Any = ...
    def __init__(self, target: Any, *order_by: Any) -> None: ...
    def self_group(self, against: Optional[Any] = ...): ...
    def get_children(self, **kwargs: Any): ...

class ExcludeConstraint(ColumnCollectionConstraint):
    __visit_name__: str = ...
    where: Any = ...
    create_drop_stringify_dialect: str = ...
    operators: Any = ...
    using: Any = ...
    ops: Any = ...
    def __init__(self, *elements: Any, **kw: Any) -> None: ...
    def copy(self, target_table: Optional[Any] = ..., **kw: Any): ...

def array_agg(*arg: Any, **kw: Any): ...



import sys

if sys.version_info >= (3, 0):
    import asyncio as asyncio  # noqa
    from typing import Any
    from typing import Callable
    from typing import Coroutine
    from typing import TypeVar

    _T = TypeVar("_T")
    _AAL = TypeVar("_AAL", bound=AsyncAdaptedLock)
    def await_only(awaitable: Coroutine[_T, Any, Any]) -> _T: ...
    def await_fallback(awaitable: Coroutine[_T, Any, Any]) -> _T: ...
    async def greenlet_spawn(
        fn: Callable[..., _T],
        *args: Any,
        _require_await: bool = ...,
        **kwargs: Any,
    ) -> _T: ...
    class AsyncAdaptedLock:
        mutex: Any = ...
        def __init__(self) -> None: ...
        def __enter__(self: _AAL) -> _AAL: ...
        def __exit__(self, *arg: Any, **kw: Any) -> None: ...



import abc
from typing import Any

from .interfaces import _DBAPIConnection
from .interfaces import Dialect
from ..util import ABC

class ConnectionCharacteristic(ABC, metaclass=abc.ABCMeta):
    transactional: bool = ...
    @abc.abstractmethod
    def reset_characteristic(
        self, dialect: Dialect, dbapi_conn: _DBAPIConnection
    ) -> None: ...
    @abc.abstractmethod
    def set_characteristic(
        self, dialect: Dialect, dbapi_conn: _DBAPIConnection, value: Any
    ) -> None: ...
    @abc.abstractmethod
    def get_characteristic(
        self, dialect: Dialect, dbapi_conn: _DBAPIConnection
    ) -> Any: ...

class IsolationLevelCharacteristic(ConnectionCharacteristic):
    transactional: bool = ...
    def reset_characteristic(
        self, dialect: Dialect, dbapi_conn: _DBAPIConnection
    ) -> None: ...
    def set_characteristic(
        self, dialect: Dialect, dbapi_conn: _DBAPIConnection, value: Any
    ) -> None: ...
    def get_characteristic(
        self, dialect: Dialect, dbapi_conn: _DBAPIConnection
    ) -> Any: ...



from .array import All as All
from .array import Any as Any
from .array import ARRAY as ARRAY
from .array import array as array
from .base import BIGINT as BIGINT
from .base import BIT as BIT
from .base import BOOLEAN as BOOLEAN
from .base import BYTEA as BYTEA
from .base import CHAR as CHAR
from .base import CIDR as CIDR
from .base import CreateEnumType as CreateEnumType
from .base import DATE as DATE
from .base import DOUBLE_PRECISION as DOUBLE_PRECISION
from .base import DropEnumType as DropEnumType
from .base import ENUM as ENUM
from .base import FLOAT as FLOAT
from .base import INET as INET
from .base import INTEGER as INTEGER
from .base import INTERVAL as INTERVAL
from .base import MACADDR as MACADDR
from .base import MONEY as MONEY
from .base import NUMERIC as NUMERIC
from .base import OID as OID
from .base import REAL as REAL
from .base import REGCLASS as REGCLASS
from .base import SMALLINT as SMALLINT
from .base import TEXT as TEXT
from .base import TIME as TIME
from .base import TIMESTAMP as TIMESTAMP
from .base import TSVECTOR as TSVECTOR
from .base import UUID as UUID
from .base import VARCHAR as VARCHAR
from .dml import Insert as Insert
from .dml import insert as insert
from .ext import aggregate_order_by as aggregate_order_by
from .ext import array_agg as array_agg
from .ext import ExcludeConstraint as ExcludeConstraint
from .hstore import HSTORE as HSTORE
from .hstore import hstore as hstore
from .json import JSON as JSON
from .json import JSONB as JSONB
from .ranges import DATERANGE as DATERANGE
from .ranges import INT4RANGE as INT4RANGE
from .ranges import INT8RANGE as INT8RANGE
from .ranges import NUMRANGE as NUMRANGE
from .ranges import TSRANGE as TSRANGE
from .ranges import TSTZRANGE as TSTZRANGE

dialect: Any



from typing import Any
from typing import Optional

from .json import JSON as JSON
from .json import JSONIndexType as JSONIndexType
from .json import JSONPathType as JSONPathType
from ... import exc as exc
from ... import Identity as Identity
from ... import Sequence as Sequence
from ... import sql as sql
from ... import types as sqltypes
from ... import util as util
from ...engine import default as default
from ...engine import reflection as reflection
from ...sql import coercions as coercions
from ...sql import compiler as compiler
from ...sql import elements as elements
from ...sql import expression as expression
from ...sql import func as func
from ...sql import quoted_name as quoted_name
from ...sql import roles as roles
from ...types import BIGINT as BIGINT
from ...types import BINARY as BINARY
from ...types import CHAR as CHAR
from ...types import DATE as DATE
from ...types import DATETIME as DATETIME
from ...types import DECIMAL as DECIMAL
from ...types import FLOAT as FLOAT
from ...types import INTEGER as INTEGER
from ...types import NCHAR as NCHAR
from ...types import NUMERIC as NUMERIC
from ...types import NVARCHAR as NVARCHAR
from ...types import SMALLINT as SMALLINT
from ...types import TEXT as TEXT
from ...types import VARCHAR as VARCHAR
from ...util import compat as compat
from ...util import update_wrapper as update_wrapper
from ...util.langhelpers import public_factory as public_factory

MS_2017_VERSION: Any
MS_2016_VERSION: Any
MS_2014_VERSION: Any
MS_2012_VERSION: Any
MS_2008_VERSION: Any
MS_2005_VERSION: Any
MS_2000_VERSION: Any
RESERVED_WORDS: Any

class REAL(sqltypes.REAL):
    __visit_name__: str = ...
    def __init__(self, **kw: Any) -> None: ...

class TINYINT(sqltypes.Integer):
    __visit_name__: str = ...

class _MSDate(sqltypes.Date):
    def bind_processor(self, dialect: Any): ...
    def result_processor(self, dialect: Any, coltype: Any): ...

class TIME(sqltypes.TIME):
    precision: Any = ...
    def __init__(
        self, precision: Optional[Any] = ..., **kwargs: Any
    ) -> None: ...
    def bind_processor(self, dialect: Any): ...
    def result_processor(self, dialect: Any, coltype: Any): ...

class _DateTimeBase:
    def bind_processor(self, dialect: Any): ...

class _MSDateTime(_DateTimeBase, sqltypes.DateTime): ...

class SMALLDATETIME(_DateTimeBase, sqltypes.DateTime):
    __visit_name__: str = ...

class DATETIME2(_DateTimeBase, sqltypes.DateTime):
    __visit_name__: str = ...
    precision: Any = ...
    def __init__(self, precision: Optional[Any] = ..., **kw: Any) -> None: ...

class DATETIMEOFFSET(_DateTimeBase, sqltypes.DateTime):
    __visit_name__: str = ...
    precision: Any = ...
    def __init__(self, precision: Optional[Any] = ..., **kw: Any) -> None: ...

class _UnicodeLiteral:
    def literal_processor(self, dialect: Any): ...

class _MSUnicode(_UnicodeLiteral, sqltypes.Unicode): ...
class _MSUnicodeText(_UnicodeLiteral, sqltypes.UnicodeText): ...

class TIMESTAMP(sqltypes._Binary):
    __visit_name__: str = ...
    length: Any = ...
    convert_int: Any = ...
    def __init__(self, convert_int: bool = ...) -> None: ...
    def result_processor(self, dialect: Any, coltype: Any): ...

class ROWVERSION(TIMESTAMP):
    __visit_name__: str = ...

class NTEXT(sqltypes.UnicodeText):
    __visit_name__: str = ...

class VARBINARY(sqltypes.VARBINARY, sqltypes.LargeBinary):
    __visit_name__: str = ...

class IMAGE(sqltypes.LargeBinary):
    __visit_name__: str = ...

class XML(sqltypes.Text):
    __visit_name__: str = ...

class BIT(sqltypes.TypeEngine):
    __visit_name__: str = ...

class MONEY(sqltypes.TypeEngine):
    __visit_name__: str = ...

class SMALLMONEY(sqltypes.TypeEngine):
    __visit_name__: str = ...

class UNIQUEIDENTIFIER(sqltypes.TypeEngine):
    __visit_name__: str = ...

class SQL_VARIANT(sqltypes.TypeEngine):
    __visit_name__: str = ...

class TryCast(sql.elements.Cast):
    __visit_name__: str = ...
    stringify_dialect: str = ...
    def __init__(self, *arg: Any, **kw: Any) -> None: ...

try_cast: Any
MSDateTime: Any
MSDate: Any
MSReal = REAL
MSTinyInteger = TINYINT
MSTime = TIME
MSSmallDateTime = SMALLDATETIME
MSDateTime2 = DATETIME2
MSDateTimeOffset = DATETIMEOFFSET
MSText = TEXT
MSNText = NTEXT
MSString = VARCHAR
MSNVarchar = NVARCHAR
MSChar = CHAR
MSNChar = NCHAR
MSBinary = BINARY
MSVarBinary = VARBINARY
MSImage = IMAGE
MSBit = BIT
MSMoney = MONEY
MSSmallMoney = SMALLMONEY
MSUniqueIdentifier = UNIQUEIDENTIFIER
MSVariant = SQL_VARIANT
ischema_names: Any

class MSTypeCompiler(compiler.GenericTypeCompiler):
    def visit_FLOAT(self, type_: Any, **kw: Any): ...
    def visit_TINYINT(self, type_: Any, **kw: Any): ...
    def visit_DATETIMEOFFSET(self, type_: Any, **kw: Any): ...
    def visit_TIME(self, type_: Any, **kw: Any): ...
    def visit_TIMESTAMP(self, type_: Any, **kw: Any): ...
    def visit_ROWVERSION(self, type_: Any, **kw: Any): ...
    def visit_DATETIME2(self, type_: Any, **kw: Any): ...
    def visit_SMALLDATETIME(self, type_: Any, **kw: Any): ...
    def visit_unicode(self, type_: Any, **kw: Any): ...
    def visit_text(self, type_: Any, **kw: Any): ...
    def visit_unicode_text(self, type_: Any, **kw: Any): ...
    def visit_NTEXT(self, type_: Any, **kw: Any): ...
    def visit_TEXT(self, type_: Any, **kw: Any): ...
    def visit_VARCHAR(self, type_: Any, **kw: Any): ...
    def visit_CHAR(self, type_: Any, **kw: Any): ...
    def visit_NCHAR(self, type_: Any, **kw: Any): ...
    def visit_NVARCHAR(self, type_: Any, **kw: Any): ...
    def visit_date(self, type_: Any, **kw: Any): ...
    def visit_time(self, type_: Any, **kw: Any): ...
    def visit_large_binary(self, type_: Any, **kw: Any): ...
    def visit_IMAGE(self, type_: Any, **kw: Any): ...
    def visit_XML(self, type_: Any, **kw: Any): ...
    def visit_VARBINARY(self, type_: Any, **kw: Any): ...
    def visit_boolean(self, type_: Any, **kw: Any): ...
    def visit_BIT(self, type_: Any, **kw: Any): ...
    def visit_JSON(self, type_: Any, **kw: Any): ...
    def visit_MONEY(self, type_: Any, **kw: Any): ...
    def visit_SMALLMONEY(self, type_: Any, **kw: Any): ...
    def visit_UNIQUEIDENTIFIER(self, type_: Any, **kw: Any): ...
    def visit_SQL_VARIANT(self, type_: Any, **kw: Any): ...

class MSExecutionContext(default.DefaultExecutionContext):
    def pre_exec(self) -> None: ...
    cursor_fetch_strategy: Any = ...
    def post_exec(self) -> None: ...
    def get_lastrowid(self): ...
    @property
    def rowcount(self): ...
    def handle_dbapi_exception(self, e: Any) -> None: ...
    def get_result_cursor_strategy(self, result: Any): ...
    def fire_sequence(self, seq: Any, type_: Any): ...
    def get_insert_default(self, column: Any): ...

class MSSQLCompiler(compiler.SQLCompiler):
    returning_precedes_values: bool = ...
    extract_map: Any = ...
    tablealiases: Any = ...
    def __init__(self, *args: Any, **kwargs: Any) -> None: ...
    def visit_now_func(self, fn: Any, **kw: Any): ...
    def visit_current_date_func(self, fn: Any, **kw: Any): ...
    def visit_length_func(self, fn: Any, **kw: Any): ...
    def visit_char_length_func(self, fn: Any, **kw: Any): ...
    def visit_concat_op_binary(
        self, binary: Any, operator: Any, **kw: Any
    ): ...
    def visit_true(self, expr: Any, **kw: Any): ...
    def visit_false(self, expr: Any, **kw: Any): ...
    def visit_match_op_binary(self, binary: Any, operator: Any, **kw: Any): ...
    def get_select_precolumns(self, select: Any, **kw: Any): ...
    def get_from_hint_text(self, table: Any, text: Any): ...
    def get_crud_hint_text(self, table: Any, text: Any): ...
    def fetch_clause(self, cs: Any, **kwargs: Any): ...
    def limit_clause(self, cs: Any, **kwargs: Any): ...
    def visit_try_cast(self, element: Any, **kw: Any): ...
    def translate_select_structure(self, select_stmt: Any, **kwargs: Any): ...
    def visit_table(
        self,
        table: Any,
        mssql_aliased: bool = ...,
        iscrud: bool = ...,
        **kwargs: Any,
    ): ...
    def visit_alias(self, alias: Any, **kw: Any): ...
    def visit_column(
        self, column: Any, add_to_result_map: Optional[Any] = ..., **kw: Any
    ): ...
    def visit_extract(self, extract: Any, **kw: Any): ...
    def visit_savepoint(self, savepoint_stmt: Any): ...
    def visit_rollback_to_savepoint(self, savepoint_stmt: Any): ...
    def visit_binary(self, binary: Any, **kwargs: Any): ...
    def returning_clause(self, stmt: Any, returning_cols: Any): ...
    def get_cte_preamble(self, recursive: Any): ...
    def label_select_column(self, select: Any, column: Any, asfrom: Any): ...
    def for_update_clause(self, select: Any, **kw: Any): ...
    def order_by_clause(self, select: Any, **kw: Any): ...
    def update_from_clause(
        self,
        update_stmt: Any,
        from_table: Any,
        extra_froms: Any,
        from_hints: Any,
        **kw: Any,
    ): ...
    def delete_table_clause(
        self, delete_stmt: Any, from_table: Any, extra_froms: Any
    ): ...
    def delete_extra_from_clause(
        self,
        delete_stmt: Any,
        from_table: Any,
        extra_froms: Any,
        from_hints: Any,
        **kw: Any,
    ): ...
    def visit_empty_set_expr(self, type_: Any): ...
    def visit_is_distinct_from_binary(
        self, binary: Any, operator: Any, **kw: Any
    ): ...
    def visit_is_not_distinct_from_binary(
        self, binary: Any, operator: Any, **kw: Any
    ): ...
    def visit_json_getitem_op_binary(
        self, binary: Any, operator: Any, **kw: Any
    ): ...
    def visit_json_path_getitem_op_binary(
        self, binary: Any, operator: Any, **kw: Any
    ): ...
    def visit_sequence(self, seq: Any, **kw: Any): ...

class MSSQLStrictCompiler(MSSQLCompiler):
    ansi_bind_rules: bool = ...
    def visit_in_op_binary(self, binary: Any, operator: Any, **kw: Any): ...
    def visit_not_in_op_binary(
        self, binary: Any, operator: Any, **kw: Any
    ): ...
    def render_literal_value(self, value: Any, type_: Any): ...

class MSDDLCompiler(compiler.DDLCompiler):
    def get_column_specification(self, column: Any, **kwargs: Any): ...
    def visit_create_index(self, create: Any, include_schema: bool = ...): ...
    def visit_drop_index(self, drop: Any): ...
    def visit_primary_key_constraint(self, constraint: Any): ...
    def visit_unique_constraint(self, constraint: Any): ...
    def visit_computed_column(self, generated: Any): ...
    def visit_create_sequence(self, create: Any, **kw: Any): ...
    def visit_identity_column(self, identity: Any, **kw: Any): ...

class MSIdentifierPreparer(compiler.IdentifierPreparer):
    reserved_words: Any = ...
    def __init__(self, dialect: Any) -> None: ...
    def quote_schema(self, schema: Any, force: Optional[Any] = ...): ...

class MSDialect(default.DefaultDialect):
    name: str = ...
    supports_default_values: bool = ...
    supports_empty_insert: bool = ...
    execution_ctx_cls: Any = ...
    use_scope_identity: bool = ...
    max_identifier_length: int = ...
    schema_name: str = ...
    implicit_returning: bool = ...
    full_returning: bool = ...
    colspecs: Any = ...
    engine_config_types: Any = ...
    ischema_names: Any = ...
    supports_sequences: bool = ...
    sequences_optional: bool = ...
    default_sequence_base: int = ...
    supports_native_boolean: bool = ...
    non_native_boolean_check_constraint: bool = ...
    supports_unicode_binds: bool = ...
    postfetch_lastrowid: bool = ...
    legacy_schema_aliasing: bool = ...
    server_version_info: Any = ...
    statement_compiler: Any = ...
    ddl_compiler: Any = ...
    type_compiler: Any = ...
    preparer: Any = ...
    construct_arguments: Any = ...
    query_timeout: Any = ...
    deprecate_large_types: Any = ...
    isolation_level: Any = ...
    def __init__(
        self,
        query_timeout: Optional[Any] = ...,
        use_scope_identity: bool = ...,
        schema_name: str = ...,
        isolation_level: Optional[Any] = ...,
        deprecate_large_types: Optional[Any] = ...,
        json_serializer: Optional[Any] = ...,
        json_deserializer: Optional[Any] = ...,
        legacy_schema_aliasing: Optional[Any] = ...,
        **opts: Any,
    ) -> None: ...
    def do_savepoint(self, connection: Any, name: Any) -> None: ...
    def do_release_savepoint(self, connection: Any, name: Any) -> None: ...
    def set_isolation_level(self, connection: Any, level: Any) -> None: ...
    def get_isolation_level(self, connection: Any): ...
    def initialize(self, connection: Any) -> None: ...
    def on_connect(self): ...
    def has_table(
        self,
        connection: Any,
        tablename: Any,
        dbname: Any,
        owner: Any,
        schema: Any,
    ): ...
    def has_sequence(
        self,
        connection: Any,
        sequencename: Any,
        dbname: Any,
        owner: Any,
        schema: Any,
    ): ...
    def get_sequence_names(
        self, connection: Any, dbname: Any, owner: Any, schema: Any, **kw: Any
    ): ...
    def get_schema_names(self, connection: Any, **kw: Any): ...
    def get_table_names(
        self, connection: Any, dbname: Any, owner: Any, schema: Any, **kw: Any
    ): ...
    def get_view_names(
        self, connection: Any, dbname: Any, owner: Any, schema: Any, **kw: Any
    ): ...
    def get_indexes(
        self,
        connection: Any,
        tablename: Any,
        dbname: Any,
        owner: Any,
        schema: Any,
        **kw: Any,
    ): ...
    def get_view_definition(
        self,
        connection: Any,
        viewname: Any,
        dbname: Any,
        owner: Any,
        schema: Any,
        **kw: Any,
    ): ...
    def get_columns(
        self,
        connection: Any,
        tablename: Any,
        dbname: Any,
        owner: Any,
        schema: Any,
        **kw: Any,
    ): ...
    def get_pk_constraint(
        self,
        connection: Any,
        tablename: Any,
        dbname: Any,
        owner: Any,
        schema: Any,
        **kw: Any,
    ): ...
    def get_foreign_keys(
        self,
        connection: Any,
        tablename: Any,
        dbname: Any,
        owner: Any,
        schema: Any,
        **kw: Any,
    ): ...



import sys
from threading import Condition
from threading import RLock
from typing import Any
from typing import Deque
from typing import Generic
from typing import NoReturn
from typing import Optional
from typing import TypeVar

_T = TypeVar("_T")

class Empty(Exception): ...
class Full(Exception): ...

class Queue(Generic[_T]):
    maxsize: int = ...
    mutex: RLock = ...
    not_empty: Condition = ...
    not_full: Condition = ...
    use_lifo: bool = ...
    queue: Deque[_T] = ...
    def __init__(self, maxsize: int = ..., use_lifo: bool = ...) -> None: ...
    def qsize(self) -> int: ...
    def empty(self) -> bool: ...
    def full(self) -> bool: ...
    def put(
        self, item: _T, block: bool = ..., timeout: Optional[float] = ...
    ) -> None: ...
    def put_nowait(self, item: _T) -> None: ...
    def get(self, block: bool = ..., timeout: Optional[float] = ...) -> _T: ...
    def get_nowait(self) -> _T: ...

class AsyncAdaptedQueue(Generic[_T]):
    use_lifo: bool = ...
    maxsize: int = ...
    @staticmethod
    def await_(awaitable: Any) -> Any: ...
    def __init__(self, maxsize: int = ..., use_lifo: bool = ...) -> None: ...
    if sys.version_info >= (3, 0):
        def empty(self) -> bool: ...
        def full(self) -> bool: ...
        def qsize(self) -> int: ...
        def put_nowait(self, item: _T) -> None: ...
        async def put(
            self, item: _T, block: bool = ..., timeout: Optional[float] = ...
        ) -> None: ...
        def get_nowait(self) -> _T: ...
        async def get(
            self, block: bool = ..., timeout: Optional[float] = ...
        ) -> _T: ...
    else:
        def empty(self) -> NoReturn: ...
        def full(self) -> NoReturn: ...
        def qsize(self) -> NoReturn: ...
        def put_nowait(self, item: Any) -> NoReturn: ...
        def put(
            self, item: Any, block: bool = ..., timeout: Optional[float] = ...
        ) -> NoReturn: ...
        def get_nowait(self) -> NoReturn: ...
        def get(
            self, block: bool = ..., timeout: Optional[float] = ...
        ) -> NoReturn: ...

class FallbackAsyncAdaptedQueue(AsyncAdaptedQueue[_T]):
    @staticmethod
    def await_(awaitable: Any) -> Any: ...



from typing import Any
from typing import Optional

from . import attributes as attributes
from . import interfaces as interfaces
from .descriptor_props import SynonymProperty as SynonymProperty
from .properties import ColumnProperty as ColumnProperty
from .util import class_mapper as class_mapper
from .. import exc as exc
from .. import inspection as inspection
from .. import util as util

def add_class(classname: Any, cls: Any, decl_class_registry: Any) -> None: ...
def remove_class(
    classname: Any, cls: Any, decl_class_registry: Any
) -> None: ...

class _MultipleClassMarker:
    on_remove: Any = ...
    contents: Any = ...
    def __init__(
        self, classes: Any, on_remove: Optional[Any] = ...
    ) -> None: ...
    def remove_item(self, cls: Any) -> None: ...
    def __iter__(self) -> Any: ...
    def attempt_get(self, path: Any, key: Any): ...
    def add_item(self, item: Any) -> None: ...

class _ModuleMarker:
    parent: Any = ...
    name: Any = ...
    contents: Any = ...
    mod_ns: Any = ...
    path: Any = ...
    def __init__(self, name: Any, parent: Any) -> None: ...
    def __contains__(self, name: Any): ...
    def __getitem__(self, name: Any): ...
    def resolve_attr(self, key: Any): ...
    def get_module(self, name: Any): ...
    def add_class(self, name: Any, cls: Any): ...
    def remove_class(self, name: Any, cls: Any) -> None: ...

class _ModNS:
    def __init__(self, parent: Any) -> None: ...
    def __getattr__(self, key: Any): ...

class _GetColumns:
    cls: Any = ...
    def __init__(self, cls: Any) -> None: ...
    def __getattr__(self, key: Any): ...

class _GetTable:
    key: Any = ...
    metadata: Any = ...
    def __init__(self, key: Any, metadata: Any) -> None: ...
    def __getattr__(self, key: Any): ...

class _class_resolver:
    cls: Any = ...
    prop: Any = ...
    arg: Any = ...
    fallback: Any = ...
    favor_tables: Any = ...
    def __init__(
        self,
        cls: Any,
        prop: Any,
        fallback: Any,
        arg: Any,
        favor_tables: bool = ...,
    ) -> None: ...
    def __call__(self): ...



import collections
from collections import namedtuple as namedtuple  # noqa
import contextlib
import operator
import sys
import threading as threading  # noqa
from types import CodeType
from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterator
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Type
from typing import Union

if sys.version_info >= (3, 0):
    from _typeshed import SupportsLessThan

    _SortKeyFunction = Callable[[Any], SupportsLessThan]
else:
    _SortKeyFunction = Callable[[Any], Any]

py38: bool
py37: bool
py3k: bool
py2k: bool
pypy: bool
cpython: bool
win32: bool
osx: bool
arm: bool
has_refcount_gc: bool

contextmanager = contextlib.contextmanager
dottedgetter = operator.attrgetter
namedtuple = collections.namedtuple
next = next

class FullArgSpec(NamedTuple):
    args: List[str]
    varargs: Optional[str]
    varkw: Optional[str]
    defaults: Optional[Tuple[Any, ...]]
    kwonlyargs: List[str]
    kwonlydefaults: Optional[Dict[str, Any]]
    annotations: Dict[str, Any]

class nullcontext:
    enter_result: Optional[Any] = ...
    def __init__(self, enter_result: Optional[Any] = ...) -> None: ...
    def __enter__(self) -> Optional[Any]: ...
    def __exit__(self, *excinfo: Any) -> None: ...

def inspect_getfullargspec(func: Any) -> FullArgSpec: ...

string_types: Tuple[Type[Any], ...]
binary_types: Tuple[Type[Any], ...]
int_types: Tuple[Type[Any], ...]

if sys.version_info >= (3, 0):
    import builtins
    import configparser
    import itertools
    import pickle as pickle  # noqa

    from functools import reduce as reduce  # noqa
    from io import BytesIO as _byte_buffer
    from io import StringIO as StringIO  # noqa
    from itertools import zip_longest as zip_longest  # noqa
    from time import perf_counter as perf_counter  # noqa
    from urllib.parse import (
        quote_plus as quote_plus,
        unquote_plus as unquote_plus,
        parse_qsl as parse_qsl,
        quote as quote,
        unquote as unquote,
    )

    binary_type = bytes
    text_type = str
    iterbytes = iter
    long_type = int

    itertools_filterfalse = itertools.filterfalse
    itertools_filter = filter
    itertools_imap = map

    exec_ = builtins.exec
    import_ = builtins.__import__
    print_ = builtins.print

    byte_buffer = _byte_buffer
    def b64decode(x: text_type) -> binary_type: ...
    def b64encode(x: binary_type) -> text_type: ...
    def cmp(__a: Any, __b: Any) -> int: ...
    def u(s: text_type) -> text_type: ...
    def ue(s: text_type) -> text_type: ...
    from typing import TYPE_CHECKING as TYPE_CHECKING  # noqa

    callable = callable

    from abc import ABC as ABC  # noqa
    import collections.abc

    collections_abc = collections.abc
else:
    import base64
    import ConfigParser as configparser  # noqa
    import itertools
    from typing import Mapping

    from StringIO import StringIO as StringIO  # noqa
    from cStringIO import StringIO as _byte_buffer
    from itertools import izip_longest as _zip_longest
    from time import clock as _perf_counter
    from urllib import quote as quote  # noqa
    from urllib import quote_plus as quote_plus  # noqa
    from urllib import unquote as unquote  # noqa
    from urllib import unquote_plus as unquote_plus  # noqa
    from urlparse import parse_qsl as parse_qsl  # noqa

    from abc import ABCMeta
    class ABC(object):
        __metaclass__ = ABCMeta
    import pickle as pickle  # noqa

    binary_type = str
    text_type = unicode  # noqa: F821
    long_type = long  # noqa: F821

    callable = callable
    cmp = cmp
    reduce = reduce

    b64encode = base64.b64encode
    b64decode = base64.b64decode

    itertools_filterfalse = itertools.ifilterfalse
    itertools_filter = itertools.ifilter
    itertools_imap = itertools.imap

    byte_buffer = _byte_buffer
    perf_counter = _perf_counter
    zip_longest = _zip_longest
    def exec_(
        __source: Union[unicode, str, CodeType],  # noqa: F821
        __globals: Optional[Dict[str, Any]],
        __locals: Optional[Mapping[str, Any]],
    ) -> None: ...
    def iterbytes(__buf: binary_type) -> Iterator[int]: ...
    def import_(*args: Any) -> Any: ...
    def print_(*args: Any, **kwargs: Any) -> None: ...
    def u(s: binary_type) -> text_type: ...
    def ue(s: binary_type) -> text_type: ...
    def safe_bytestring(text: Any) -> Any: ...
    TYPE_CHECKING: bool

    collections_abc = collections

def b(s: Any) -> Any: ...
def decode_backslashreplace(text: binary_type, encoding: str) -> text_type: ...
def raise_(
    exception: Any,
    with_traceback: Optional[Any] = ...,
    replace_context: Optional[Any] = ...,
    from_: bool = ...,
) -> None: ...
def inspect_formatargspec(
    args: Any,
    varargs: Optional[Any] = ...,
    varkw: Optional[Any] = ...,
    defaults: Optional[Any] = ...,
    kwonlyargs: Any = ...,
    kwonlydefaults: Any = ...,
    annotations: Any = ...,
    formatarg: Any = ...,
    formatvarargs: Any = ...,
    formatvarkw: Any = ...,
    formatvalue: Any = ...,
    formatreturns: Any = ...,
    formatannotation: Any = ...,
) -> str: ...

if sys.version_info >= (3, 7):
    from dataclasses import Field
    def dataclass_fields(cls: Any) -> Sequence[Field[Any]]: ...
    def local_dataclass_fields(cls: Any) -> List[Field[Any]]: ...

else:
    def dataclass_fields(cls: Any) -> Sequence[Any]: ...
    def local_dataclass_fields(cls: Any) -> List[Any]: ...

def raise_from_cause(
    exception: Any, exc_info: Optional[Any] = ...
) -> None: ...
def reraise(
    tp: Any, value: Any, tb: Optional[Any] = ..., cause: Optional[Any] = ...
) -> None: ...
def with_metaclass(meta: Any, *bases: Any, **kw: Any) -> Type[Any]: ...

if sys.version_info >= (3, 0):
    from datetime import timezone as timezone  # noqa
else:
    from datetime import datetime
    from datetime import timedelta
    from datetime import tzinfo
    class timezone(tzinfo):
        def __init__(self, offset: timedelta) -> None: ...
        def __eq__(self, other: Any) -> Any: ...
        def __hash__(self) -> int: ...
        def utcoffset(self, dt: Optional[datetime]) -> timedelta: ...
        def tzname(self, dt: Optional[datetime]) -> str: ...
        def dst(self, dt: Optional[datetime]) -> None: ...
        def fromutc(self, dt: datetime) -> datetime: ...



# fmt: off
from .engine import create_engine as create_engine
from .engine import create_mock_engine as create_mock_engine
from .engine import engine_from_config as engine_from_config
from .inspection import inspect as inspect
from .schema import BLANK_SCHEMA as BLANK_SCHEMA
from .schema import CheckConstraint as CheckConstraint
from .schema import Column as Column
from .schema import ColumnDefault as ColumnDefault
from .schema import Computed as Computed
from .schema import Constraint as Constraint
from .schema import DDL as DDL
from .schema import DefaultClause as DefaultClause
from .schema import FetchedValue as FetchedValue
from .schema import ForeignKey as ForeignKey
from .schema import ForeignKeyConstraint as ForeignKeyConstraint
from .schema import Identity as Identity
from .schema import Index as Index
from .schema import MetaData as MetaData
from .schema import PrimaryKeyConstraint as PrimaryKeyConstraint
from .schema import Sequence as Sequence
from .schema import Table as Table
from .schema import ThreadLocalMetaData as ThreadLocalMetaData
from .schema import UniqueConstraint as UniqueConstraint
from .sql import alias as alias
from .sql import all_ as all_
from .sql import and_ as and_
from .sql import any_ as any_
from .sql import asc as asc
from .sql import between as between
from .sql import bindparam as bindparam
from .sql import case as case
from .sql import cast as cast
from .sql import collate as collate
from .sql import column as column
from .sql import delete as delete
from .sql import desc as desc
from .sql import distinct as distinct
from .sql import except_ as except_
from .sql import except_all as except_all
from .sql import exists as exists
from .sql import extract as extract
from .sql import false as false
from .sql import func as func
from .sql import funcfilter as funcfilter
from .sql import insert as insert
from .sql import intersect as intersect
from .sql import intersect_all as intersect_all
from .sql import join as join
from .sql import LABEL_STYLE_DEFAULT as LABEL_STYLE_DEFAULT
from .sql import LABEL_STYLE_DISAMBIGUATE_ONLY as LABEL_STYLE_DISAMBIGUATE_ONLY
from .sql import LABEL_STYLE_NONE as LABEL_STYLE_NONE
from .sql import LABEL_STYLE_TABLENAME_PLUS_COL as LABEL_STYLE_TABLENAME_PLUS_COL
from .sql import lambda_stmt as lambda_stmt
from .sql import lateral as lateral
from .sql import literal as literal
from .sql import literal_column as literal_column
from .sql import modifier as modifier
from .sql import not_ as not_
from .sql import null as null
from .sql import nulls_first as nulls_first
from .sql import nulls_last as nulls_last
from .sql import nullsfirst as nullsfirst
from .sql import nullslast as nullslast
from .sql import or_ as or_
from .sql import outerjoin as outerjoin
from .sql import outparam as outparam
from .sql import over as over
from .sql import select as select
from .sql import subquery as subquery
from .sql import table as table
from .sql import tablesample as tablesample
from .sql import text as text
from .sql import true as true
from .sql import tuple_ as tuple_
from .sql import type_coerce as type_coerce
from .sql import union as union
from .sql import union_all as union_all
from .sql import update as update
from .sql import values as values
from .sql import within_group as within_group
from .types import ARRAY as ARRAY
from .types import BIGINT as BIGINT
from .types import BigInteger as BigInteger
from .types import BINARY as BINARY
from .types import BLOB as BLOB
from .types import BOOLEAN as BOOLEAN
from .types import Boolean as Boolean
from .types import CHAR as CHAR
from .types import CLOB as CLOB
from .types import DATE as DATE
from .types import Date as Date
from .types import DATETIME as DATETIME
from .types import DateTime as DateTime
from .types import DECIMAL as DECIMAL
from .types import Enum as Enum
from .types import FLOAT as FLOAT
from .types import Float as Float
from .types import INT as INT
from .types import INTEGER as INTEGER
from .types import Integer as Integer
from .types import Interval as Interval
from .types import JSON as JSON
from .types import LargeBinary as LargeBinary
from .types import NCHAR as NCHAR
from .types import NUMERIC as NUMERIC
from .types import Numeric as Numeric
from .types import NVARCHAR as NVARCHAR
from .types import PickleType as PickleType
from .types import REAL as REAL
from .types import SMALLINT as SMALLINT
from .types import SmallInteger as SmallInteger
from .types import String as String
from .types import TEXT as TEXT
from .types import Text as Text
from .types import TIME as TIME
from .types import Time as Time
from .types import TIMESTAMP as TIMESTAMP
from .types import TypeDecorator as TypeDecorator
from .types import Unicode as Unicode
from .types import UnicodeText as UnicodeText
from .types import VARBINARY as VARBINARY
from .types import VARCHAR as VARCHAR



from typing import Any

from .impl import QueuePool as QueuePool
from .. import util as util
from ..util import threading as threading

proxies: Any

def manage(module: Any, **params: Any): ...
def clear_managers() -> None: ...

class _DBProxy:
    module: Any = ...
    kw: Any = ...
    poolclass: Any = ...
    pools: Any = ...
    def __init__(
        self, module: Any, poolclass: Any = ..., **kw: Any
    ) -> None: ...
    def close(self) -> None: ...
    def __del__(self) -> None: ...
    def __getattr__(self, key: Any): ...
    def get_pool(self, *args: Any, **kw: Any): ...
    def connect(self, *args: Any, **kw: Any): ...
    def dispose(self, *args: Any, **kw: Any) -> None: ...



from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import overload
from typing import Type
from typing import Union

from typing_extensions import Literal

from .base import Engine
from .url import URL
from ..future import Engine as FutureEngine
from ..pool import Pool

_Debug = Literal["debug"]
_IsolationLevel = Literal[
    "SERIALIZABLE",
    "REPEATABLE READ",
    "READ COMMITTED",
    "READ UNCOMMITTED",
    "AUTOCOMMIT",
]
_ParamStyle = Literal["qmark", "numeric", "named", "format", "pyformat"]
_ResetOnReturn = Literal["rollback", "commit"]

@overload
def create_engine(
    url: Union[str, URL],
    *,
    case_sensitive: bool = ...,
    connect_args: Dict[Any, Any] = ...,
    convert_unicode: bool = ...,
    creator: Callable[..., Any] = ...,
    echo: Union[bool, _Debug] = ...,
    echo_pool: Union[bool, _Debug] = ...,
    enable_from_linting: bool = ...,
    encoding: str = ...,
    execution_options: Dict[Any, Any] = ...,
    future: Literal[True],
    hide_parameters: bool = ...,
    implicit_returning: bool = ...,
    isolation_level: _IsolationLevel = ...,
    json_deserializer: Callable[..., Any] = ...,
    json_serializer: Callable[..., Any] = ...,
    label_length: Optional[int] = ...,
    listeners: Any = ...,
    logging_name: str = ...,
    max_identifier_length: Optional[int] = ...,
    max_overflow: int = ...,
    module: Optional[Any] = ...,
    paramstyle: Optional[_ParamStyle] = ...,
    pool: Optional[Pool] = ...,
    poolclass: Optional[Type[Pool]] = ...,
    pool_logging_name: str = ...,
    pool_pre_ping: bool = ...,
    pool_size: int = ...,
    pool_recycle: int = ...,
    pool_reset_on_return: Optional[_ResetOnReturn] = ...,
    pool_timeout: float = ...,
    pool_use_lifo: bool = ...,
    plugins: List[str] = ...,
    query_cache_size: int = ...,
    **kwargs: Any,
) -> FutureEngine: ...
@overload
def create_engine(
    url: Union[str, URL],
    *,
    case_sensitive: bool = ...,
    connect_args: Dict[Any, Any] = ...,
    convert_unicode: bool = ...,
    creator: Callable[..., Any] = ...,
    echo: Union[bool, _Debug] = ...,
    echo_pool: Union[bool, _Debug] = ...,
    enable_from_linting: bool = ...,
    encoding: str = ...,
    execution_options: Dict[Any, Any] = ...,
    future: Literal[False] = ...,
    hide_parameters: bool = ...,
    implicit_returning: bool = ...,
    isolation_level: _IsolationLevel = ...,
    json_deserializer: Callable[..., Any] = ...,
    json_serializer: Callable[..., Any] = ...,
    label_length: Optional[int] = ...,
    listeners: Any = ...,
    logging_name: str = ...,
    max_identifier_length: Optional[int] = ...,
    max_overflow: int = ...,
    module: Optional[Any] = ...,
    paramstyle: Optional[_ParamStyle] = ...,
    pool: Optional[Pool] = ...,
    poolclass: Optional[Type[Pool]] = ...,
    pool_logging_name: str = ...,
    pool_pre_ping: bool = ...,
    pool_size: int = ...,
    pool_recycle: int = ...,
    pool_reset_on_return: Optional[_ResetOnReturn] = ...,
    pool_timeout: float = ...,
    pool_use_lifo: bool = ...,
    plugins: List[str] = ...,
    query_cache_size: int = ...,
    **kwargs: Any,
) -> Engine: ...
@overload
def create_engine(
    url: Union[str, URL],
    *,
    case_sensitive: bool = ...,
    connect_args: Dict[Any, Any] = ...,
    convert_unicode: bool = ...,
    creator: Callable[..., Any] = ...,
    echo: Union[bool, _Debug] = ...,
    echo_pool: Union[bool, _Debug] = ...,
    enable_from_linting: bool = ...,
    encoding: str = ...,
    execution_options: Dict[Any, Any] = ...,
    future: Union[bool],
    hide_parameters: bool = ...,
    implicit_returning: bool = ...,
    isolation_level: _IsolationLevel = ...,
    json_deserializer: Callable[..., Any] = ...,
    json_serializer: Callable[..., Any] = ...,
    label_length: Optional[int] = ...,
    listeners: Any = ...,
    logging_name: str = ...,
    max_identifier_length: Optional[int] = ...,
    max_overflow: int = ...,
    module: Optional[Any] = ...,
    paramstyle: Optional[_ParamStyle] = ...,
    pool: Optional[Pool] = ...,
    poolclass: Optional[Type[Pool]] = ...,
    pool_logging_name: str = ...,
    pool_pre_ping: bool = ...,
    pool_size: int = ...,
    pool_recycle: int = ...,
    pool_reset_on_return: Optional[_ResetOnReturn] = ...,
    pool_timeout: float = ...,
    pool_use_lifo: bool = ...,
    plugins: List[str] = ...,
    query_cache_size: int = ...,
    **kwargs: Any,
) -> Union[Engine, FutureEngine]: ...
def engine_from_config(
    configuration: Any, prefix: str = ..., **kwargs: Any
) -> Engine: ...



import abc
from typing import Any
from typing import Generator
from typing import TypeVar

_TStartableContext = TypeVar("_TStartableContext", bound=StartableContext)

class StartableContext(abc.ABC, metaclass=abc.ABCMeta):
    @abc.abstractmethod
    async def start(self: _TStartableContext) -> _TStartableContext: ...
    def __await__(
        self: _TStartableContext,
    ) -> Generator[Any, None, _TStartableContext]: ...
    async def __aenter__(self: _TStartableContext) -> _TStartableContext: ...
    @abc.abstractmethod
    async def __aexit__(
        self, type_: Any, value: Any, traceback: Any
    ) -> None: ...

class ProxyComparable:
    def __hash__(self) -> int: ...
    def __eq__(self, other: Any) -> bool: ...
    def __ne__(self, other: Any) -> bool: ...



from typing import Any
from typing import Optional

from .. import util as util
from ..orm import attributes as attributes
from ..orm import interfaces as interfaces
from ..sql import elements as elements

HYBRID_METHOD: Any
HYBRID_PROPERTY: Any

class hybrid_method(interfaces.InspectionAttrInfo):
    is_attribute: bool = ...
    extension_type: Any = ...
    func: Any = ...
    def __init__(self, func: Any, expr: Optional[Any] = ...) -> None: ...
    def __get__(self, instance: Any, owner: Any): ...
    expr: Any = ...
    def expression(self, expr: Any): ...

class hybrid_property(interfaces.InspectionAttrInfo):
    is_attribute: bool = ...
    extension_type: Any = ...
    fget: Any = ...
    fset: Any = ...
    fdel: Any = ...
    expr: Any = ...
    custom_comparator: Any = ...
    update_expr: Any = ...
    def __init__(
        self,
        fget: Any,
        fset: Optional[Any] = ...,
        fdel: Optional[Any] = ...,
        expr: Optional[Any] = ...,
        custom_comparator: Optional[Any] = ...,
        update_expr: Optional[Any] = ...,
    ) -> None: ...
    def __get__(self, instance: Any, owner: Any): ...
    def __set__(self, instance: Any, value: Any) -> None: ...
    def __delete__(self, instance: Any) -> None: ...
    @property
    def overrides(self): ...
    def getter(self, fget: Any): ...
    def setter(self, fset: Any): ...
    def deleter(self, fdel: Any): ...
    def expression(self, expr: Any): ...
    def comparator(self, comparator: Any): ...
    def update_expression(self, meth: Any): ...

class Comparator(interfaces.PropComparator):
    property: Any = ...
    expression: Any = ...
    def __init__(self, expression: Any) -> None: ...
    def __clause_element__(self): ...
    def adapt_to_entity(self, adapt_to_entity: Any): ...

class ExprComparator(Comparator):
    cls: Any = ...
    expression: Any = ...
    hybrid: Any = ...
    def __init__(self, cls: Any, expression: Any, hybrid: Any) -> None: ...
    def __getattr__(self, key: Any): ...
    @property
    def info(self): ...
    @property
    def property(self): ...
    def operate(self, op: Any, *other: Any, **kwargs: Any): ...
    def reverse_operate(self, op: Any, other: Any, **kwargs: Any): ...
